import argparse
import ast
import os
import shutil
from typing import Any, Callable, Dict, FrozenSet, List, Optional, Set, Tuple

import dlt
from dlt.common.pipeline import get_dlt_pipelines_dir
from dlt.common.schema import Schema
from dlt.common.storages.configuration import TSchemaFileFormat
from dlt.common.time import ensure_pendulum_datetime_non_utc
from dlt.common.typing import TAnyDateTime, TFun
from dlt.common.configuration.container import Container
from dlt.common.configuration.providers.provider import ConfigProvider
from dlt.common.configuration.resolve import resolve_configuration
from dlt.common.configuration.specs.pluggable_run_context import (
    PluggableRunContext,
    RunContextBase,
    ProfilesRunContext,
)
from dlt.common.configuration.specs.runtime_configuration import RuntimeConfiguration
from dlt.common.reflection.utils import set_ast_parents
from dlt.common.runtime import run_context
from dlt.common.runtime.exec_info import get_plus_version
from dlt.common.runtime.telemetry import with_telemetry
from dlt.common.storages.file_storage import FileStorage
from dlt.common.versioned_state import json_decode_state

from dlt.pipeline.pipeline import Pipeline
from dlt.pipeline.trace import get_trace_file_path
from dlt.reflection.script_visitor import PipelineScriptVisitor

from dlt._workspace.cli.exceptions import CliCommandException, CliCommandInnerException
from dlt._workspace.cli import echo as fmt
from dlt._workspace.helpers.dashboard.typing import TPipelineListItem
from dlt._workspace.profile import BUILT_IN_PROFILES, read_profile_pin
from dlt._workspace.typing import (
    ProviderInfo,
    ProviderLocationInfo,
    TCurrentProfileInfo,
    TLocationInfo,
    TLocationScope,
    TProfileInfo,
    TProviderInfo,
    TSchemaExport,
    TToolkitIndexEntry,
    TWorkspaceInfo,
)


REQUIREMENTS_TXT = "requirements.txt"
PYPROJECT_TOML = "pyproject.toml"
GITHUB_WORKFLOWS_DIR = os.path.join(".github", "workflows")

DEFAULT_MCP_FEATURES: FrozenSet[str] = frozenset(
    {"workspace", "pipeline", "toolkit", "secrets", "context"}
)
"""Default MCP feature set. Defined here (not in server.py) so argparsers can
reference it without importing fastmcp."""
AIRFLOW_DAGS_FOLDER = os.path.join("dags")
AIRFLOW_BUILD_FOLDER = os.path.join("build")
MODULE_INIT = "__init__.py"
DATETIME_FORMAT = "YYYY-MM-DD HH:mm:ss"


def get_pipeline_trace_mtime(pipelines_dir: str, pipeline_name: str) -> float:
    """Get mtime of the trace saved by pipeline, which approximates run time"""
    trace_file = get_trace_file_path(pipelines_dir, pipeline_name)
    if os.path.isfile(trace_file):
        return os.path.getmtime(trace_file)
    return 0


def _get_pipeline_initial_cwd(pipelines_dir: str, pipeline_name: str) -> Optional[str]:
    """Read initial_cwd from a pipeline's local state without attaching."""
    state_path = os.path.join(pipelines_dir, pipeline_name, Pipeline.STATE_FILE)
    try:
        with open(state_path, encoding="utf-8") as f:
            state = json_decode_state(f.read())
        local: Dict[str, Any] = state.get("_local", {})
        return local.get("initial_cwd")  # type: ignore[no-any-return]
    except (OSError, ValueError, KeyError):
        return None


def list_local_pipelines(
    pipelines_dir: str = None,
    sort_by_trace: bool = True,
    additional_pipelines: List[str] = None,
    run_dir: Optional[str] = None,
) -> Tuple[str, List[TPipelineListItem]]:
    """Get the local pipelines directory and the list of pipeline names in it.

    Args:
        pipelines_dir: The local pipelines directory. Defaults to get_dlt_pipelines_dir().
        sort_by_trace: Whether to sort the pipelines by the latest timestamp of trace.
        additional_pipelines: Extra pipeline names to include in the list.
        run_dir: When set, only return pipelines whose initial_cwd matches this path.
    """
    pipelines_dir = pipelines_dir or get_dlt_pipelines_dir()
    storage = FileStorage(pipelines_dir)

    try:
        pipelines = storage.list_folder_dirs(".", to_root=False)
    except Exception:
        pipelines = []

    if additional_pipelines:
        for pipeline in additional_pipelines:
            if pipeline and pipeline not in pipelines:
                pipelines.append(pipeline)

    if run_dir:
        abs_project_dir = os.path.abspath(run_dir)
        pipelines = [
            p for p in pipelines if _get_pipeline_initial_cwd(pipelines_dir, p) == abs_project_dir
        ]

    # check last trace timestamp and create dict
    pipelines_with_timestamps: List[TPipelineListItem] = []
    for pipeline in pipelines:
        pipelines_with_timestamps.append(
            {"name": pipeline, "timestamp": get_pipeline_trace_mtime(pipelines_dir, pipeline)}
        )

    if sort_by_trace:
        pipelines_with_timestamps.sort(key=lambda x: x["timestamp"], reverse=True)

    return pipelines_dir, pipelines_with_timestamps


def date_from_timestamp_with_ago(
    timestamp: TAnyDateTime, datetime_format: str = DATETIME_FORMAT
) -> str:
    """Return a date with ago section"""
    if not timestamp or timestamp == 0:
        return "never"
    timestamp = ensure_pendulum_datetime_non_utc(timestamp)
    time_formatted = timestamp.format(datetime_format)
    ago = timestamp.diff_for_humans()
    return f"{ago} ({time_formatted})"


def display_run_context_info() -> None:
    run_context = dlt.current.run_context()
    if isinstance(run_context, ProfilesRunContext):
        if run_context.default_profile != run_context.profile:
            # print warning
            fmt.echo(
                "Profile `%s` is active."
                % (fmt.style(run_context.profile, fg="yellow", reset=True),),
                err=True,
            )


def make_mcp_run_flags(default_port: int = 8000) -> argparse.ArgumentParser:
    """Build a parent parser with the shared MCP run flags (--stdio, --sse, --port, --features).

    Returns an ArgumentParser with `add_help=False` suitable for use as a
    `parents=[...]` entry.  Does **not** require `fastmcp` to be installed.
    """
    flags = argparse.ArgumentParser(add_help=False)
    flags.add_argument("--stdio", action="store_true", help="Use stdio transport mode")
    flags.add_argument(
        "--sse",
        action="store_true",
        help="Use legacy SSE transport instead of streamable-http",
    )
    flags.add_argument(
        "--port",
        type=int,
        default=default_port,
        help="Port for the MCP server (default: %d)" % default_port,
    )
    defaults = sorted(DEFAULT_MCP_FEATURES)
    flags.add_argument(
        "--features",
        nargs="*",
        default=None,
        help=(
            "MCP features to enable/disable. Default: %s. "
            "Use +name to add, -name to remove "
            "(e.g. --features=-secrets,+context)"
            % ", ".join(defaults)
        ),
    )
    return flags


def add_mcp_arg_parser(subparsers: Any, description: str, help_str: str, default_port: int) -> None:
    try:
        from dlt._workspace.mcp.server import WorkspaceMCP  # noqa: F401
    except ImportError:
        return

    flags = make_mcp_run_flags(default_port)
    subparsers.add_parser(
        "mcp",
        description=description,
        help=help_str,
        parents=[flags],
    )


def _may_safe_delete_local(run_context: RunContextBase, deleted_dir_type: str) -> bool:
    deleted_dir = getattr(run_context, deleted_dir_type)
    deleted_abs = os.path.abspath(deleted_dir)
    run_dir_abs = os.path.abspath(run_context.run_dir)
    settings_dir_abs = os.path.abspath(run_context.settings_dir)

    # never allow deleting run_dir or settings_dir themselves
    for ctx_abs, label in (
        (run_dir_abs, "run dir (workspace root)"),
        (settings_dir_abs, "settings dir"),
    ):
        if deleted_abs == ctx_abs:
            fmt.error(
                f"{deleted_dir_type} `{deleted_dir}` is the same as {label} and cannot be deleted"
            )
            return False

    # ensure deleted directory is inside run_dir
    try:
        common = os.path.commonpath([deleted_abs, run_dir_abs])
    except ValueError:
        # occurs when paths are on different drives on windows
        common = ""
    if common != run_dir_abs:
        fmt.error(
            f"{deleted_dir_type} `{deleted_dir}` is not within run dir (workspace root) and cannot"
            " be deleted"
        )
        return False

    return True


def _wipe_dir(
    run_context: RunContextBase, dir_attr: str, echo_template: str, recreate_dirs: bool = True
) -> None:
    """Echo, safely wipe and optionally recreate a directory from run context.

    Args:
        run_context: Current run context.
        dir_attr: Attribute name on the run context that holds the directory path, eg. "local_dir".
        echo_template: Template used to echo the action to the user. Must contain a single %s placeholder for the styled path.
        recreate_dirs: when True, recreate the directory after deletion.
    """
    dir_path = getattr(run_context, dir_attr, None)
    if not dir_path:
        raise CliCommandException()

    # ensure we never attempt to operate on run_dir or settings_dir
    if not _may_safe_delete_local(run_context, dir_attr):
        raise CliCommandException()

    # show relative path to the user when shorter
    display_dir = os.path.relpath(dir_path, ".")
    if len(display_dir) > len(dir_path):
        display_dir = dir_path

    fmt.echo(echo_template % fmt.style(display_dir, fg="yellow"))

    if os.path.exists(dir_path):
        shutil.rmtree(dir_path, onerror=FileStorage.rmtree_del_ro)

    if recreate_dirs:
        os.makedirs(dir_path, exist_ok=True)


def check_delete_local_data(run_context: RunContextBase, skip_data_dir: bool) -> List[str]:
    """Display paths to be deleted and ask for confirmation.

    Args:
        run_context: current run context.
        skip_data_dir: when True, do not include the data_dir in the deletion set.

    Returns:
        A list of run_context attribute names that should be deleted. Empty list if user cancels.

    Raises:
        CliCommandException: if context is invalid or deletion is not safe.
    """
    # ensure profiles context
    if not isinstance(run_context, ProfilesRunContext):
        fmt.error("Cannot delete local data for a context without profiles")
        raise CliCommandException()

    attrs: list[str] = ["local_dir"]
    if not skip_data_dir:
        attrs.append("data_dir")

    # ensure we never attempt to operate on run_dir or settings_dir
    for attr in attrs:
        if not _may_safe_delete_local(run_context, attr):
            raise CliCommandException()

    # display relative paths to run_dir
    fmt.echo("The following dirs will be deleted:")
    for attr in attrs:
        dir_path = getattr(run_context, attr)
        display_dir = os.path.relpath(dir_path, run_context.run_dir)
        if attr == "local_dir":
            template = "- %s (locally loaded data)"
        elif attr == "data_dir":
            template = "- %s (pipeline working folders)"
        else:
            raise ValueError(attr)

        fmt.echo(template % fmt.style(display_dir, fg="yellow", reset=True))

    # ask for confirmation
    if not fmt.confirm("Do you want to proceed?", default=False):
        return []

    return attrs


def delete_local_data(
    run_context: RunContextBase, dir_attrs: List[str], recreate_dirs: bool = True
) -> None:
    """Delete local data directories after explicit confirmation.

    Args:
        run_context: current run context.
        dir_attrs: A list of run_context attribute names that should be deleted.
        recreate_dirs: when True, recreate directories after deletion.
    """

    # delete selected directories
    for attr in dir_attrs:
        _wipe_dir(run_context, attr, "Deleting %s", recreate_dirs)


def parse_init_script(
    command: str, script_source: str, init_script_name: str
) -> PipelineScriptVisitor:
    # parse the script first
    tree = ast.parse(source=script_source)
    set_ast_parents(tree)
    visitor = PipelineScriptVisitor(script_source)
    visitor.visit_passes(tree)
    if len(visitor.mod_aliases) == 0:
        raise CliCommandInnerException(
            command,
            f"The pipeline script {init_script_name} does not import dlt and does not seem to run"
            " any pipelines",
        )

    return visitor


def ensure_git_command(command: str) -> None:
    try:
        import git
    except ImportError as imp_ex:
        if "Bad git executable" not in str(imp_ex):
            raise
        raise CliCommandInnerException(
            command,
            "'git' command is not available. Install and setup git with the following the guide %s"
            % "https://docs.github.com/en/get-started/quickstart/set-up-git",
            imp_ex,
        ) from imp_ex


def track_command(
    command: str, track_before: bool, *args: str, **kwargs: str
) -> Callable[[TFun], TFun]:
    """Return a telemetry decorator for CLI commands.

    Wraps a function with anonymous telemetry tracking using `with_telemetry`. Depending on
    `track_before`, emits an event either before execution or after execution with success
    information.

    Success semantics:
    - if the wrapped function returns an int, 0 is treated as success; other values as failure.
    - for non-int returns, success is True unless an exception is raised.

    Args:
        command: event/command name to report.
        track_before: if True, emit a single event before calling the function. if False,
            emit a single event after the call, including success state.
        *args: names of parameters from the decorated function whose values should be included
            in the event properties.
        **kwargs: additional key-value pairs to include in the event properties.

    Returns:
        a decorator that applies telemetry tracking to the decorated function.
    """
    return with_telemetry("command", command, track_before, *args, **kwargs)


def get_telemetry_status() -> bool:
    c = resolve_configuration(RuntimeConfiguration())
    return c.dlthub_telemetry


def make_dlt_settings_path(path: str = None) -> str:
    """Returns path to file in dlt settings folder. Returns settings folder if path not specified."""
    ctx = run_context.active()
    if not path:
        return ctx.settings_dir
    return ctx.get_setting(path)


def get_provider_locations() -> List[ProviderInfo]:
    """Return structured info about all config providers and their locations.

    Works with both RunContext (OSS, no profiles) and WorkspaceRunContext
    (profile-aware). Always reflects the currently active profile.
    """
    ctx_plug = Container()[PluggableRunContext]
    ctx = ctx_plug.context
    providers = ctx_plug.providers.providers

    settings_dir = os.path.abspath(ctx.settings_dir)
    global_dir = os.path.abspath(ctx.global_dir)

    result: List[ProviderInfo] = []
    for provider in providers:
        profile: Optional[str] = getattr(provider, "_profile", None)
        present_set = set(os.path.abspath(p) for p in provider.present_locations)

        loc_infos: List[ProviderLocationInfo] = []
        for path in provider.locations:
            abs_path = os.path.abspath(path)
            is_present = abs_path in present_set

            # determine scope: project if under settings_dir, global otherwise
            if abs_path.startswith(settings_dir + os.sep) or abs_path == settings_dir:
                scope: TLocationScope = "project"
            elif abs_path.startswith(global_dir + os.sep) or abs_path == global_dir:
                scope = "global"
            else:
                scope = "global"

            # determine if this specific location is profile-scoped
            profile_name: Optional[str] = None
            if profile and os.path.basename(path).startswith(f"{profile}."):
                profile_name = profile

            loc_infos.append(ProviderLocationInfo(path, is_present, scope, profile_name))

        result.append(ProviderInfo(provider, loc_infos))
    return result


def fetch_profiles_list() -> List[TProfileInfo]:
    """Return all available profiles with their status flags.

    Works with ProfilesRunContext (workspace). Returns an empty list for OSS RunContext.
    """
    ctx = Container()[PluggableRunContext].context
    if not isinstance(ctx, ProfilesRunContext):
        return []

    pinned = read_profile_pin(ctx)
    current = ctx.profile
    configured = set(ctx.configured_profiles())

    return [
        TProfileInfo(
            name=name,
            description=BUILT_IN_PROFILES.get(name, "custom profile"),
            is_current=name == current,
            is_pinned=name == pinned,
            is_configured=name in configured,
        )
        for name in ctx.available_profiles()
    ]


def fetch_workspace_info() -> TWorkspaceInfo:
    """Return workspace information as a structured dict.

    Works with both OSS RunContext (no profiles) and WorkspaceRunContext.
    Always includes all provider locations (verbose mode).
    """
    from dlt._workspace.cli.ai.utils import load_toolkits_index

    ctx = Container()[PluggableRunContext].context

    # profile info — only when profiles are available
    profile_info: Optional[TCurrentProfileInfo] = None
    configured_profiles: List[str] = []
    if isinstance(ctx, ProfilesRunContext):
        configured_profiles = ctx.configured_profiles()
        profile_info = TCurrentProfileInfo(
            name=ctx.profile,
            description="",
            is_current=True,
            is_pinned=ctx.profile == read_profile_pin(ctx),
            is_configured=ctx.profile in configured_profiles,
            data_dir=ctx.data_dir,
            local_dir=ctx.local_dir,
        )

    # workspace name — only meaningful for WorkspaceRunContext
    name: Optional[str] = None
    if isinstance(ctx, ProfilesRunContext):
        name = ctx.name

    # provider locations — always verbose (all locations)
    providers: List[TProviderInfo] = []
    for info in get_provider_locations():
        providers.append(
            TProviderInfo(
                name=info.provider.name,
                is_empty=info.provider.is_empty,
                locations=[
                    TLocationInfo(
                        path=loc.path,
                        present=loc.present,
                        scope=loc.scope,
                        profile_name=loc.profile_name,
                    )
                    for loc in info.locations
                ],
            )
        )

    # dlt and dlthub versions
    plus_version = get_plus_version()
    dlthub_version: Optional[str] = plus_version["version"] if plus_version else None

    # initialized: config.toml exists in settings dir
    initialized = os.path.isfile(make_dlt_settings_path("config.toml"))

    # installed toolkits from local index
    installed_toolkits: Dict[str, TToolkitIndexEntry] = load_toolkits_index()

    return TWorkspaceInfo(
        name=name,
        run_dir=ctx.run_dir,
        settings_dir=ctx.settings_dir,
        global_dir=ctx.global_dir,
        profile=profile_info,
        configured_profiles=configured_profiles,
        providers=providers,
        dlt_version=dlt.__version__,
        dlthub_version=dlthub_version,
        initialized=initialized,
        installed_toolkits=installed_toolkits,
    )


def fetch_schema_export(
    schema: Schema,
    format_: TSchemaFileFormat = "yaml",
    remove_defaults: bool = True,
    hide_columns: bool = False,
) -> TSchemaExport:
    """Export a schema object in the requested format."""
    if format_ == "json":
        content = schema.to_pretty_json(remove_defaults=remove_defaults)
    elif format_ == "yaml":
        content = schema.to_pretty_yaml(remove_defaults=remove_defaults)
    elif format_ == "dbml":
        content = schema.to_dbml()
    elif format_ == "dot":
        content = schema.to_dot()
    elif format_ == "mermaid":
        content = schema.to_mermaid(hide_columns=hide_columns)
    else:
        content = schema.to_pretty_yaml(remove_defaults=remove_defaults)
    return TSchemaExport(
        schema_name=schema.name,
        format_=format_,
        content=content,
    )



import re
import abc
import os
import yaml
import ast

from yaml import Dumper
from itertools import chain
from typing import List, Optional, Sequence, Tuple, Any, Dict

try:
    import cron_descriptor
except ImportError:
    cron_descriptor = None

import dlt

from dlt.common.libs import git
from dlt.common.configuration.exceptions import LookupTrace, ConfigFieldMissingException
from dlt.common.configuration.providers import (
    CONFIG_TOML,
    EnvironProvider,
    StringTomlProvider,
)
from dlt.common.libs.git import get_origin, get_repo, Repo
from dlt.common.typing import StrAny
from dlt.common.reflection.utils import evaluate_node_literal
from dlt.common.pipeline import LoadInfo, TPipelineState, get_dlt_repos_dir
from dlt.common.storages import FileStorage
from dlt.common.utils import set_working_dir

from dlt.pipeline.pipeline import Pipeline
from dlt.pipeline.trace import PipelineTrace
from dlt.pipeline.configuration import get_default_pipeline_name
from dlt.reflection import names as n
from dlt.reflection.script_visitor import PipelineScriptVisitor

from dlt._workspace.cli import utils
from dlt._workspace.cli import echo as fmt
from dlt._workspace.cli.exceptions import CliCommandInnerException, PipelineWasNotRun

GITHUB_URL = "https://github.com/"


def get_schedule_description(schedule: str) -> Optional[Any]:
    try:
        return None if schedule is None else cron_descriptor.get_description(schedule)
    except Exception as ex:
        raise ValueError(f"Error when parsing schedule '{schedule}': {str(ex)}")


class BaseDeployment(abc.ABC):
    def __init__(
        self,
        pipeline_script_path: str,
        location: str,
        branch: Optional[str] = None,
    ):
        self.pipeline_script_path = pipeline_script_path
        self.repo_location = location
        self.branch = branch

        self.pipelines_dir: Optional[str] = None
        self.pipeline_name: Optional[str] = None

        self.deployment_method: str
        self.repo: Repo
        self.repo_storage: FileStorage
        self.origin: str
        self.repo_pipeline_script_path: str
        self.pipeline_script: Any
        self.template_storage: FileStorage
        self.working_directory: str
        self.state: TPipelineState

        self.env_prov = EnvironProvider()
        self.envs: List[LookupTrace] = []
        self.secret_envs: List[LookupTrace] = []
        self.artifacts: Dict[str, Any] = {}

    def _prepare_deployment(self) -> None:
        self.repo_storage = FileStorage(str(self.repo.working_dir))
        # make sure the repo has origin
        self.origin = self._get_origin()
        # convert to path relative to repo
        self.repo_pipeline_script_path = self.repo_storage.from_wd_to_relative_path(
            self.pipeline_script_path
        )
        # load a pipeline script and extract full_refresh and pipelines_dir args
        self.pipeline_script = self.repo_storage.load(self.repo_pipeline_script_path)
        fmt.echo(
            "Looking up the deployment template scripts in %s...\n" % fmt.bold(self.repo_location)
        )
        self.template_storage = git.get_fresh_repo_files(
            self.repo_location, get_dlt_repos_dir(), branch=self.branch
        )
        self.working_directory = os.path.split(self.pipeline_script_path)[0]

    def _get_origin(self) -> str:
        try:
            origin = get_origin(self.repo)
            if "github.com" not in origin:
                raise CliCommandInnerException(
                    "deploy",
                    f"Your current repository origin is not set to github but to {origin}.\nYou"
                    " must change it to be able to run the pipelines with github actions:"
                    " https://docs.github.com/en/get-started/getting-started-with-git/managing-remote-repositories",
                )
        except ValueError:
            raise CliCommandInnerException(
                "deploy",
                "Your current repository has no origin set. Please set it up to be able to run the"
                " pipelines with github actions:"
                " https://docs.github.com/en/get-started/importing-your-projects-to-github/importing-source-code-to-github/adding-locally-hosted-code-to-github",
            )

        return origin

    def run_deployment(self) -> None:
        with get_repo(self.pipeline_script_path) as repo:
            self.repo = repo
            self._prepare_deployment()
            # go through all once launched pipelines
            visitors = get_visitors(self.pipeline_script, self.pipeline_script_path)
            possible_pipelines = parse_pipeline_info(visitors)
            pipeline_name: str = None
            pipelines_dir: str = None

            uniq_possible_pipelines = {t[0]: t for t in possible_pipelines}
            if len(uniq_possible_pipelines) == 1:
                pipeline_name, pipelines_dir = possible_pipelines[0]
            elif len(uniq_possible_pipelines) > 1:
                choices = list(uniq_possible_pipelines.keys())
                choices_str = "".join([str(i + 1) for i in range(len(choices))])
                choices_selection = [f"{idx+1}-{name}" for idx, name in enumerate(choices)]
                sel = fmt.prompt(
                    "Several pipelines found in script, please select one: "
                    + ", ".join(choices_selection),
                    choices=choices_str,
                )
                pipeline_name, pipelines_dir = uniq_possible_pipelines[choices[int(sel) - 1]]

            if pipelines_dir:
                self.pipelines_dir = os.path.abspath(pipelines_dir)
            if pipeline_name:
                self.pipeline_name = pipeline_name

            # change the working dir to the script working dir
            with set_working_dir(self.working_directory):
                # use script name to derive pipeline name
                if not self.pipeline_name:
                    self.pipeline_name = dlt.config.get("pipeline_name")
                    if not self.pipeline_name:
                        self.pipeline_name = get_default_pipeline_name(self.pipeline_script_path)
                        fmt.warning(
                            f"Using default pipeline name {self.pipeline_name}. The pipeline name"
                            " is not passed as argument to dlt.pipeline nor configured via config"
                            " provides ie. config.toml"
                        )
                # fmt.echo("Generating deployment for pipeline %s" % fmt.bold(self.pipeline_name))

                # attach to pipeline name, get state and trace
                pipeline = dlt.attach(
                    pipeline_name=self.pipeline_name, pipelines_dir=self.pipelines_dir
                )
                self.state, trace = get_state_and_trace(pipeline)
                self._update_envs(trace)

            self._generate_workflow()
            self._echo_instructions()
            self._make_modification()

    def _update_envs(self, trace: PipelineTrace) -> None:
        # add destination name and dataset name to env
        self.envs = [
            # LookupTrace(self.env_prov.name, (), "destination_name", self.state["destination"]),
            # LookupTrace(self.env_prov.name, (), "dataset_name", self.state["dataset_name"])
        ]

        for resolved_value in trace.resolved_config_values:
            if resolved_value.is_secret_hint:
                # generate special forms for all secrets
                self.secret_envs.append(
                    LookupTrace(
                        self.env_prov.name,
                        tuple(resolved_value.sections),
                        resolved_value.key,
                        resolved_value.value,
                    )
                )
                # fmt.echo(f"{resolved_value.key}:{resolved_value.value}{type(resolved_value.value)} in {resolved_value.sections} is SECRET")
            else:
                # move all config values that are not in config.toml into env
                if resolved_value.provider_name != CONFIG_TOML:
                    self.envs.append(
                        LookupTrace(
                            self.env_prov.name,
                            tuple(resolved_value.sections),
                            resolved_value.key,
                            resolved_value.value,
                        )
                    )
                    # fmt.echo(f"{resolved_value.key} in {resolved_value.sections} moved to CONFIG")

    def _echo_secrets(self) -> None:
        display_info = False
        for s_v in self.secret_envs:
            fmt.secho("Name:", fg="green")
            fmt.echo(fmt.bold(self.env_prov.get_key_name(s_v.key, *s_v.sections)))
            try:
                fmt.secho("Secret:", fg="green")
                fmt.echo(self._lookup_secret_value(s_v))
            except ConfigFieldMissingException:
                fmt.secho("please set me up!", fg="red")
                display_info = True
            fmt.echo()
        if display_info:
            self._display_missing_secret_info()
            fmt.echo()

    def _echo_secrets_toml(self) -> None:
        display_info = False
        toml_provider = StringTomlProvider("")
        for s_v in self.secret_envs:
            try:
                secret_value = self._lookup_secret_value(s_v)
            except ConfigFieldMissingException:
                secret_value = "please set me up!"
                display_info = True
            toml_provider.set_value(s_v.key, secret_value, None, *s_v.sections)
        for s_v in self.envs:
            toml_provider.set_value(s_v.key, s_v.value, None, *s_v.sections)
        fmt.echo(toml_provider.dumps())
        if display_info:
            self._display_missing_secret_info()
            fmt.echo()

    def _display_missing_secret_info(self) -> None:
        fmt.warning(
            "We could not read and display some secrets. Starting from 1.0 version of dlt,"
            " those are not stored in the traces. Instead we are trying to read them from the"
            " available configuration ie. secrets.toml file. Please run the deploy command from"
            " the same working directory you ran your pipeline script. If you pass the"
            " credentials in code we will not be able to display them here. See"
            " https://dlthub.com/docs/general-usage/credentials/"
        )

    def _lookup_secret_value(self, trace: LookupTrace) -> Any:
        return dlt.secrets[StringTomlProvider.get_key_name(trace.key, *trace.sections)]

    def _echo_envs(self) -> None:
        for v in self.envs:
            fmt.secho("Name:", fg="green")
            fmt.echo(fmt.bold(self.env_prov.get_key_name(v.key, *v.sections)))
            fmt.secho("Value:", fg="green")
            fmt.echo(v.value)
            fmt.echo()

    @abc.abstractmethod
    def _echo_instructions(self, *args: Optional[Any]) -> Optional[Any]:
        pass

    @abc.abstractmethod
    def _generate_workflow(self, *args: Optional[Any]) -> Optional[Any]:
        pass

    @abc.abstractmethod
    def _make_modification(self) -> None:
        pass


def get_state_and_trace(pipeline: Pipeline) -> Tuple[TPipelineState, PipelineTrace]:
    # trace must exist and end with a successful loading step
    trace = pipeline.last_trace
    if trace is None or len(trace.steps) == 0:
        raise PipelineWasNotRun(
            "Pipeline run trace could not be found. Please run the pipeline at least once locally."
        )
    last_step = trace.steps[-1]
    if last_step.step_exception is not None:
        raise PipelineWasNotRun(
            "The last pipeline run ended with error. Please make sure that pipeline runs correctly"
            f" before deployment.\n{last_step.step_exception}"
        )
    if not isinstance(last_step.step_info, LoadInfo):
        raise PipelineWasNotRun(
            "The last pipeline run did not reach the load step. Please run the pipeline locally"
            " until it loads data into destination."
        )

    return pipeline.state, trace


def get_visitors(pipeline_script: str, pipeline_script_path: str) -> PipelineScriptVisitor:
    visitor = utils.parse_init_script("deploy", pipeline_script, pipeline_script_path)
    if n.RUN not in visitor.known_calls:
        raise CliCommandInnerException(
            "deploy",
            f"The pipeline script {pipeline_script_path} does not seem to run the pipeline.",
        )
    return visitor


def parse_pipeline_info(visitor: PipelineScriptVisitor) -> List[Tuple[str, Optional[str]]]:
    pipelines: List[Tuple[str, Optional[str]]] = []
    if n.PIPELINE in visitor.known_calls:
        for call_args in visitor.known_calls[n.PIPELINE]:
            pipeline_name, pipelines_dir = None, None
            # Check both full_refresh/dev_mode until full_refresh option is removed from dlt
            f_r_node = call_args.arguments.get("full_refresh") or call_args.arguments.get(
                "dev_mode"
            )
            if f_r_node:
                f_r_value = evaluate_node_literal(f_r_node)
                if f_r_value is None:
                    fmt.warning(
                        "The value of `dev_mode` in call to `dlt.pipeline` cannot be"
                        f" determined from {ast.unparse(f_r_node).strip()}. We assume that you know"
                        " what you are doing :)"
                    )
                if f_r_value is True:
                    if fmt.confirm(
                        "The value of 'dev_mode' or 'full_refresh' is set to True. Do you want to"
                        " abort to set it to False?",
                        default=True,
                    ):
                        raise CliCommandInnerException("deploy", "Please set the dev_mode to False")

            p_d_node = call_args.arguments.get("pipelines_dir")
            if p_d_node:
                pipelines_dir = evaluate_node_literal(p_d_node)
                if pipelines_dir is None:
                    raise CliCommandInnerException(
                        "deploy",
                        "The value of 'pipelines_dir' argument in call to `dlt_pipeline` cannot be"
                        f" determined from {ast.unparse(p_d_node).strip()}. Pipeline working dir"
                        " will be found. Pass it directly with --pipelines-dir option.",
                    )

            p_n_node = call_args.arguments.get("pipeline_name")
            if p_n_node:
                pipeline_name = evaluate_node_literal(p_n_node)
                if pipeline_name is None:
                    raise CliCommandInnerException(
                        "deploy",
                        "The value of 'pipeline_name' argument in call to `dlt_pipeline` cannot be"
                        f" determined from {ast.unparse(p_d_node).strip()}. Pipeline working dir"
                        " will be found. Pass it directly with --pipeline-name option.",
                    )
            pipelines.append((pipeline_name, pipelines_dir))

    return pipelines


def str_representer(dumper: yaml.Dumper, data: str) -> yaml.ScalarNode:
    # format multiline strings as blocks with the exception of placeholders
    # that will be expanded as yaml
    if len(data.splitlines()) > 1 and "{{ toYaml" not in data:  # check for multiline string
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
    return dumper.represent_scalar("tag:yaml.org,2002:str", data)


def wrap_template_str(s: str) -> str:
    return "${{ %s }}" % s


def serialize_templated_yaml(tree: StrAny) -> str:
    old_representer = Dumper.yaml_representers[str]
    try:
        yaml.add_representer(str, str_representer)
        # pretty serialize yaml
        serialized: str = yaml.dump(
            tree, allow_unicode=True, default_flow_style=False, sort_keys=False
        )
        # removes apostrophes around the template
        serialized = re.sub(r"'([\s\n]*?\${{.+?}})'", r"\1", serialized, flags=re.DOTALL)
        # print(serialized)
        # fix the new lines in templates ending }}
        serialized = re.sub(r"(\${{.+)\n.+(}})", r"\1 \2", serialized)
        return serialized
    finally:
        yaml.add_representer(str, old_representer)


def generate_pip_freeze(requirements_blacklist: List[str], requirements_file_name: str) -> str:
    try:
        import pipdeptree
    except Exception:
        fmt.warning("pipdeptree does not work anymore on this system")
        return "# please provide valid dependencies including dlt package"

    pkgs = pipdeptree.get_installed_distributions(local_only=True, user_only=False)

    # construct graph with all packages
    tree = pipdeptree.PackageDAG.from_pkgs(pkgs)
    branch_keys = {r.key for r in chain.from_iterable(tree.values())}
    # all the top level packages
    nodes = [p for p in tree.keys() if p.key not in branch_keys]

    # compute excludes to compute includes as set difference
    excludes = set(req.strip() for req in requirements_blacklist if not req.strip().startswith("#"))
    includes = set(node.project_name for node in nodes if node.project_name not in excludes)

    # prepare new filtered DAG
    tree = tree.sort()
    tree = tree.filter_nodes(includes, None)
    branch_keys = {r.key for r in chain.from_iterable(tree.values())}
    nodes = [p for p in tree.keys() if p.key not in branch_keys]

    # detect and warn on conflict
    conflicts = pipdeptree.conflicting_deps(tree)
    cycles = pipdeptree.cyclic_deps(tree)
    if conflicts:
        fmt.warning(
            "Unable to create dependencies for the github action. Please edit"
            f" {requirements_file_name} yourself"
        )
        pipdeptree.render_conflicts_text(conflicts)
        pipdeptree.render_cycles_text(cycles)
        fmt.echo()
        # do not create package because it will most probably fail
        return "# please provide valid dependencies including dlt package"

    lines = [node.render(None, False) for node in nodes]
    return "\n".join(lines)


def github_origin_to_url(origin: str, path: str) -> str:
    # repository origin must end with .git
    if origin.endswith(".git"):
        origin = origin[:-4]
    if origin.startswith("git@github.com:"):
        origin = origin[15:]

    if not origin.startswith(GITHUB_URL):
        origin = GITHUB_URL + origin
    # https://github.com/dlt-hub/data-loading-zoomcamp.git
    # git@github.com:dlt-hub/data-loading-zoomcamp.git

    # https://github.com/dlt-hub/data-loading-zoomcamp/settings/secrets/actions
    return origin + path


def ask_files_overwrite(files: Sequence[str]) -> None:
    existing = [file for file in files if os.path.exists(file)]
    if existing:
        fmt.echo("Following files will be overwritten: %s" % fmt.bold(str(existing)))
        if not fmt.confirm("Do you want to continue?", default=False):
            raise CliCommandInnerException("init", "Aborted")



from typing import Any, Dict, Type

from dlt.common.data_types.type_helpers import PY_TYPE_TO_SC_TYPE, py_type_to_sc_type
from dlt.common.data_types.typing import TDataType
from dlt.common.normalizers.json.relational import (
    DataItemNormalizer as RelationalNormalizer,
)


class DataItemNormalizer(RelationalNormalizer):
    """Normalizer that never coerces across types.

    Type mismatches always produce variant columns instead of attempting
    value conversion.
    """

    @property
    def py_type_to_sc_type_map(self) -> Dict[Type[Any], TDataType]:
        return PY_TYPE_TO_SC_TYPE

    def py_type_to_sc_type(self, t: Type[Any]) -> TDataType:
        return py_type_to_sc_type(t)

    def can_coerce_type(self, to_type: TDataType, from_type: TDataType) -> bool:
        return to_type == from_type

    def coerce_type(self, to_type: TDataType, from_type: TDataType, value: Any) -> Any:
        if to_type == from_type:
            return value
        raise ValueError(f"No cross-type coercion: cannot convert {from_type} to {to_type}")



from typing import Optional

from dlt.common.configuration import configspec
from dlt.common.destination.capabilities import TLoaderParallelismStrategy
from dlt.common.storages import LoadStorageConfiguration
from dlt.common.runners.configuration import PoolRunnerConfiguration, TPoolType


@configspec
class LoaderConfiguration(PoolRunnerConfiguration):
    workers: int = 20
    """how many parallel loads can be executed"""
    parallelism_strategy: Optional[TLoaderParallelismStrategy] = None
    """Which parallelism strategy to use at load time"""
    pool_type: TPoolType = "thread"  # mostly i/o (upload) so may be thread pool
    raise_on_failed_jobs: bool = True
    """when True, raises on terminally failed jobs immediately"""
    raise_on_max_retries: int = 5
    """When gt 0 will raise when job reaches raise_on_max_retries"""
    _load_storage_config: LoadStorageConfiguration = None
    truncate_staging_dataset: bool = False
    """If set to `True`, the staging dataset will be truncated after loading the data"""
    start_new_jobs_on_signal: bool = False
    """If set to False: will attempt to drain load pool on signal, if True: will continue loading new job"""

    def on_resolved(self) -> None:
        self.pool_type = (
            "none" if (self.workers == 1 or self.parallelism_strategy == "sequential") else "thread"
        )



import functools
import warnings
import semver
import typing
import typing_extensions

from dlt.version import __version__

VersionString = typing.Union[str, semver.Version]


class DltDeprecationWarning(DeprecationWarning):
    """A dlt specific deprecation warning.

    This warning is raised when using deprecated functionality in dlt. It provides information on when the
    deprecation was introduced and the expected version in which the corresponding functionality will be removed.

    Attributes:
        message: Description of the warning.
        since: Version in which the deprecation was introduced.
        expected_due: Version in which the corresponding functionality is expected to be removed.
    """

    def __init__(
        self,
        message: str,
        *args: typing.Any,
        since: VersionString,
        expected_due: VersionString = None,
    ) -> None:
        super().__init__(message, *args)
        self.message = message.rstrip(".")
        self.since = since if isinstance(since, semver.Version) else semver.Version.parse(since)
        if expected_due:
            expected_due = (
                expected_due
                if isinstance(expected_due, semver.Version)
                else semver.Version.parse(expected_due)
            )
        # we deprecate across major version since 1.0.0
        self.expected_due = expected_due if expected_due is not None else self.since.bump_major()

    def __str__(self) -> str:
        message = (
            f"{self.message}. Deprecated in dlt {self.since} to be removed in {self.expected_due}."
        )
        return message


class Dlt04DeprecationWarning(DltDeprecationWarning):
    V04 = semver.Version.parse("0.4.0")

    def __init__(self, message: str, *args: typing.Any, expected_due: VersionString = None) -> None:
        super().__init__(
            message, *args, since=Dlt04DeprecationWarning.V04, expected_due=expected_due
        )


class Dlt100DeprecationWarning(DltDeprecationWarning):
    V100 = semver.Version.parse("1.0.0")

    def __init__(self, message: str, *args: typing.Any, expected_due: VersionString = None) -> None:
        super().__init__(
            message, *args, since=Dlt100DeprecationWarning.V100, expected_due=expected_due
        )


# show dlt deprecations once
warnings.simplefilter("once", DltDeprecationWarning)

if typing.TYPE_CHECKING or hasattr(typing_extensions, "deprecated"):
    deprecated = typing_extensions.deprecated
else:
    # ported from typing_extensions so versions older than 4.5.x may still be used
    _T = typing.TypeVar("_T")

    def deprecated(
        __msg: str,
        *,
        category: typing.Optional[typing.Type[Warning]] = DeprecationWarning,
        stacklevel: int = 1,
    ) -> typing.Callable[[_T], _T]:
        """Indicate that a class, function or overload is deprecated.

        Usage:

            @deprecated("Use B instead")
            class A:
                pass

            @deprecated("Use g instead")
            def f():
                pass

            @overload
            @deprecated("int support is deprecated")
            def g(x: int) -> int: ...
            @overload
            def g(x: str) -> int: ...

        When this decorator is applied to an object, the type checker
        will generate a diagnostic on usage of the deprecated object.

        The warning specified by ``category`` will be emitted on use
        of deprecated objects. For functions, that happens on calls;
        for classes, on instantiation. If the ``category`` is ``None``,
        no warning is emitted. The ``stacklevel`` determines where the
        warning is emitted. If it is ``1`` (the default), the warning
        is emitted at the direct caller of the deprecated object; if it
        is higher, it is emitted further up the stack.

        The decorator sets the ``__deprecated__``
        attribute on the decorated object to the deprecation message
        passed to the decorator. If applied to an overload, the decorator
        must be after the ``@overload`` decorator for the attribute to
        exist on the overload as returned by ``get_overloads()``.

        See PEP 702 for details.

        """

        def decorator(__arg: _T) -> _T:
            if category is None:
                __arg.__deprecated__ = __msg
                return __arg
            elif isinstance(__arg, type):
                original_new = __arg.__new__
                has_init = __arg.__init__ is not object.__init__

                @functools.wraps(original_new)
                def __new__(cls, *args, **kwargs):
                    warnings.warn(__msg, category=category, stacklevel=stacklevel + 1)
                    if original_new is not object.__new__:
                        return original_new(cls, *args, **kwargs)
                    # Mirrors a similar check in object.__new__.
                    elif not has_init and (args or kwargs):
                        raise TypeError(f"{cls.__name__}() takes no arguments")
                    else:
                        return original_new(cls)

                __arg.__new__ = staticmethod(__new__)
                __arg.__deprecated__ = __new__.__deprecated__ = __msg
                return __arg
            elif callable(__arg):

                @functools.wraps(__arg)
                def wrapper(*args, **kwargs):
                    warnings.warn(__msg, category=category, stacklevel=stacklevel + 1)
                    return __arg(*args, **kwargs)

                __arg.__deprecated__ = wrapper.__deprecated__ = __msg
                return wrapper
            else:
                raise TypeError(
                    "@deprecated decorator with non-None category must be applied to "
                    f"a class or callable, not {__arg!r}"
                )

        return decorator



"""Extensible dialect capabilities for the SqlAlchemy destination.

Users can register custom DialectCapabilities subclasses to adapt the destination
for dialects that are not built-in. See register_dialect_capabilities.
"""

from typing import Any, Dict, Optional, Type

import sqlalchemy as sa  # noqa

from dlt.common.destination.capabilities import DataTypeMapper, DestinationCapabilitiesContext
from dlt.common.destination.typing import PreparedTableSchema


_GENERIC_UNDEFINED_RELATION_PATTERNS = [
    # MySQL / MariaDB
    "unknown database",
    "doesn't exist",
    "unknown table",
    # SQLite
    "no such table",
    "no such database",
    # PostgreSQL / Trino / Vertica / Exasol
    "does not exist",
    # MSSQL
    "invalid object name",
    # SAP HANA
    "invalid schema name",
    "invalid table name",
    # DB2
    "is an undefined name",
    # Apache Hive
    "table not found",
    "database does not exist",
    # Exasol (broad)
    " not found",
]
"""Generic undefined-relation patterns shared across dialects, matched against lowercased str(exception)."""

GENERIC_TERMINAL_PATTERNS = [
    "no such",
    "not found",
    "not exist",
    "unknown",
]
"""Terminal patterns for exceptions that don't match undefined-relation but are still non-transient."""


class DialectCapabilities:
    """Base class defining dialect-specific behavior for the SqlAlchemy destination.

    Subclass this to adapt the destination for a new SqlAlchemy dialect. Each method
    corresponds to an extension point:

    * adjust_capabilities -- tweak destination capabilities (identifier lengths, timestamp
      precision, sqlglot dialect, etc.)
    * type_mapper_class -- return a custom DataTypeMapper subclass for the dialect
    * adapt_table -- modify an sa.Table object before it is materialized
      (e.g. reorder columns for StarRocks)
    * is_undefined_relation -- detect "table/schema not found" errors for the dialect

    The sqlglot_dialect property maps backend names to sqlglot dialect names. Override
    it in subclasses or add entries to SQLGLOT_DIALECTS for non-obvious mappings.
    """

    SQLGLOT_DIALECTS: Dict[str, str] = {
        "postgresql": "postgres",
        "mssql": "tsql",
        "mariadb": "mysql",
        "awsathena": "athena",
        "teradatasql": "teradata",
    }
    """Backend name to sqlglot dialect name. Only entries where the two differ; others fall back to the backend name."""

    def __init__(self, backend_name: str = "") -> None:
        self._backend_name = backend_name

    @property
    def sqlglot_dialect(self) -> str:
        """The sqlglot dialect name for this backend.

        Looks up SQLGLOT_DIALECTS first, falls back to the backend name itself.
        """
        return self.SQLGLOT_DIALECTS.get(self._backend_name, self._backend_name)

    def adjust_capabilities(
        self,
        caps: DestinationCapabilitiesContext,
        dialect: sa.engine.interfaces.Dialect,
    ) -> None:
        """Adjust destination capabilities for this dialect.

        Called during adjust_capabilities on the factory. Modify caps in-place.
        """

    def type_mapper_class(self) -> Type[DataTypeMapper]:
        """Return the type mapper class for this dialect"""
        from dlt.destinations.impl.sqlalchemy.type_mapper import SqlalchemyTypeMapper

        return SqlalchemyTypeMapper

    def adapt_table(
        self,
        table: sa.Table,
        table_schema: PreparedTableSchema,
    ) -> sa.Table:
        """Modify an sa.Table object before it is created or used for loading.

        Return the (possibly modified) table. The default implementation is a no-op.
        """
        return table

    def is_undefined_relation(self, e: Exception) -> Optional[bool]:
        """Classify an exception as an undefined-relation error (or not).

        The base implementation matches generic patterns that work across many
        databases. Override in subclasses for dialect-specific error detection.

        Returns:
            True if the exception represents a missing table/schema,
            False if it is definitely not such an error, or None to fall
            through to the built-in pattern matching.
        """
        msg = str(e).lower()
        for pat in _GENERIC_UNDEFINED_RELATION_PATTERNS:
            if pat in msg:
                return True
        return None


DIALECT_CAPS_REGISTRY: Dict[str, Type[DialectCapabilities]] = {}
"""Maps dialect / backend name to the DialectCapabilities class that handles it."""


def register_dialect_capabilities(
    dialect_name: str,
    caps_class: Type[DialectCapabilities],
) -> None:
    """Register a custom DialectCapabilities for a dialect name.

    After registration the capabilities are automatically applied when the
    SqlAlchemy destination connects to a database whose backend name matches
    dialect_name.

    Args:
        dialect_name: Backend name as returned by SqlalchemyCredentials.get_backend_name()
            (e.g. "oracle", "starrocks").
        caps_class: A subclass of DialectCapabilities.

    Raises:
        ValueError: If caps_class is not a subclass of DialectCapabilities.
    """
    if not (isinstance(caps_class, type) and issubclass(caps_class, DialectCapabilities)):
        raise ValueError(
            f"caps_class must be a subclass of DialectCapabilities, got {caps_class!r}"
        )
    DIALECT_CAPS_REGISTRY[dialect_name] = caps_class


def get_dialect_capabilities(dialect_name: str) -> Optional[DialectCapabilities]:
    """Look up previously registered DialectCapabilities instance for a dialect name.
    Returns None of not found
    """
    caps_cls = DIALECT_CAPS_REGISTRY.get(dialect_name)
    if caps_cls is not None:
        return caps_cls(dialect_name)
    return None


class MysqlDialectCapabilities(DialectCapabilities):
    """Capabilities for MySQL / MariaDB."""

    def adjust_capabilities(
        self,
        caps: DestinationCapabilitiesContext,
        dialect: sa.engine.interfaces.Dialect,
    ) -> None:
        # dialect uses 255 (max length for aliases) instead of 64 (max length of identifiers)
        caps.max_identifier_length = 64
        caps.max_column_identifier_length = 64
        caps.format_datetime_literal = _format_mysql_datetime_literal
        caps.enforces_nulls_on_alter = False

    def type_mapper_class(self) -> Type[DataTypeMapper]:
        from dlt.destinations.impl.sqlalchemy.type_mapper import MysqlVariantTypeMapper

        return MysqlVariantTypeMapper


class TrinoDialectCapabilities(DialectCapabilities):
    """Capabilities for Trino."""

    def adjust_capabilities(
        self,
        caps: DestinationCapabilitiesContext,
        dialect: sa.engine.interfaces.Dialect,
    ) -> None:
        caps.timestamp_precision = 3
        caps.max_timestamp_precision = 3

    def type_mapper_class(self) -> Type[DataTypeMapper]:
        from dlt.destinations.impl.sqlalchemy.type_mapper import TrinoVariantTypeMapper

        return TrinoVariantTypeMapper


class MssqlDialectCapabilities(DialectCapabilities):
    """Capabilities for Microsoft SQL Server."""

    def type_mapper_class(self) -> Type[DataTypeMapper]:
        from dlt.destinations.impl.sqlalchemy.type_mapper import MssqlVariantTypeMapper

        return MssqlVariantTypeMapper


class OracleDialectCapabilities(DialectCapabilities):
    """Capabilities for Oracle."""

    def is_undefined_relation(self, e: Exception) -> Optional[bool]:
        msg = str(e).lower()
        # ORA-00942: table or view does not exist
        if "00942" in msg:
            return True
        return super().is_undefined_relation(e)


def _format_mysql_datetime_literal(v: Any, precision: int = 6, no_tz: bool = False) -> str:
    from dlt.common.data_writers.escape import format_datetime_literal

    return format_datetime_literal(v, precision, no_tz=True)


register_dialect_capabilities("mysql", MysqlDialectCapabilities)
register_dialect_capabilities("mariadb", MysqlDialectCapabilities)
register_dialect_capabilities("trino", TrinoDialectCapabilities)
register_dialect_capabilities("mssql", MssqlDialectCapabilities)
register_dialect_capabilities("oracle", OracleDialectCapabilities)



import datetime  # noqa: I251
import math
import re
import sys
from typing import Any, Optional, Union, overload, TypeVar, Callable  # noqa

from pendulum.parsing import (
    parse_iso8601,
    DEFAULT_OPTIONS as pendulum_options,
    _parse_common as parse_datetime_common,
)
from pendulum.tz import UTC
from pendulum import DateTime, Date, Time  # noqa: I251

from dlt.common.pendulum import create_dt, ensure_pendulum_dt, pendulum, timedelta
from dlt.common.typing import TimedeltaSeconds, TAnyDateTime
from dlt.common.warnings import deprecated

PAST_TIMESTAMP: float = 0.0
FUTURE_TIMESTAMP: float = 9999999999.0
DAY_DURATION_SEC: float = 24 * 60 * 60.0

precise_time: Callable[[], float] = None
"""A precise timer using win_precise_time library on windows and time.time on other systems"""

try:
    import win_precise_time as wpt

    precise_time = wpt.time
except ImportError:
    from time import time as _built_in_time

    precise_time = _built_in_time


class MonotonicPreciseTime:
    """Wall-clock timer guaranteed to never go backward.

    Reads wall clock on every call and tracks the highest value seen.  When the
    wall clock jumps backward (NTP step corrections, VM/WSL clock drift) the
    previous high-water mark is returned instead.

    Args:
        strictly_increasing: every call returns a value strictly greater than
            the previous one by bumping with ``math.nextafter`` when the wall
            clock does not advance. Use for callers that need unique timestamps.

    Not thread-safe. Use ``LockedMonotonicPreciseTime`` for shared instances.
    """

    def __init__(self, strictly_increasing: bool = False) -> None:
        self._last: float = precise_time()
        self._strictly_increasing = strictly_increasing

    def __call__(self) -> float:
        wall = precise_time()
        if wall > self._last:
            self._last = wall
        elif self._strictly_increasing:
            self._last = math.nextafter(self._last, math.inf)
        return self._last


class LockedMonotonicPreciseTime(MonotonicPreciseTime):
    """Thread-safe variant using a lock (one uncontended futex CAS)."""

    def __init__(self, strictly_increasing: bool = False) -> None:
        import threading

        super().__init__(strictly_increasing=strictly_increasing)
        self._lock = threading.Lock()

    def __call__(self) -> float:
        wall = precise_time()
        with self._lock:
            if wall > self._last:
                self._last = wall
            elif self._strictly_increasing:
                self._last = math.nextafter(self._last, math.inf)
            return self._last


increasing_precise_time = LockedMonotonicPreciseTime(strictly_increasing=True)


def timestamp_within(
    timestamp: float, min_exclusive: Optional[float], max_inclusive: Optional[float]
) -> bool:
    """
    check if timestamp within range uniformly treating none and range inclusiveness
    """
    return timestamp > (min_exclusive or PAST_TIMESTAMP) and timestamp <= (
        max_inclusive or FUTURE_TIMESTAMP
    )


def timestamp_before(timestamp: float, max_inclusive: Optional[float]) -> bool:
    """
    check if timestamp is before max timestamp, inclusive
    """
    return timestamp <= (max_inclusive or FUTURE_TIMESTAMP)


def parse_iso_like_datetime(value: str) -> Union[DateTime, Date, Time]:
    """Parses ISO8601 string into pendulum datetime, date or time. Preserves timezone info.

    Note: naive datetimes will be generated from string without timezone.
    """
    try:
        dtv = parse_iso8601(value)
    except ValueError:
        # fallback for formats like "2024" (year only) or "2024/01/15"
        dtv = parse_datetime_common(value, **pendulum_options)

    if isinstance(dtv, datetime.datetime):
        return ensure_pendulum_dt(dtv)
    if isinstance(dtv, datetime.date):
        return Date(dtv.year, dtv.month, dtv.day)
    if isinstance(dtv, datetime.time):
        return Time(
            dtv.hour,
            dtv.minute,
            dtv.second,
            dtv.microsecond,
            tzinfo=dtv.tzinfo,
            fold=dtv.fold,
        )
    raise ValueError(f"Interval ISO 8601 not supported: `{value}`")


def ensure_pendulum_date(value: TAnyDateTime) -> pendulum.Date:
    """Coerce a date/time value to a `pendulum.Date` object.

    UTC is assumed if the value is not timezone aware. Other timezones are shifted to UTC

    Args:
        value: The value to coerce. Can be a pendulum.DateTime, pendulum.Date, datetime, date or iso date/time str.

    Returns:
        A timezone aware pendulum.Date object.
    """
    return ensure_pendulum_datetime_non_utc(value).in_tz(UTC).date()


def ensure_pendulum_datetime_utc(value: TAnyDateTime) -> pendulum.DateTime:
    """Coerce a date/time value to a `pendulum.DateTime` object.

    UTC is assumed if the value is not timezone aware. Other timezones are shifted to UTC

    Args:
        value: The value to coerce. Can be a pendulum.DateTime, pendulum.Date, datetime, date or iso date/time str.

    Returns:
        A timezone aware pendulum.DateTime object in UTC timezone.
    """
    return ensure_pendulum_datetime_non_utc(value).in_tz(UTC)


ensure_pendulum_datetime = deprecated("Use ensure_pendulum_datetime_utc instead")(
    ensure_pendulum_datetime_utc
)


def ensure_pendulum_datetime_non_utc(value: TAnyDateTime) -> pendulum.DateTime:
    """Coerce a date/time value to a `pendulum.DateTime` object.

    Tz-awareness is preserved. Naive datetimes remain naive. Tz-aware datetimes keep their original timezone.
    Dates are converted to naive datetimes as dates are naive

    Args:
        value: The value to coerce. Can be a pendulum.DateTime, pendulum.Date, datetime, date or iso date/time str.

    Returns:
        pendulum.DateTime object that preserver original timezone
    """
    if isinstance(value, str):
        # fast path for ISO datetime strings
        try:
            dtv = parse_iso8601(value)
            if isinstance(dtv, datetime.datetime):
                return ensure_pendulum_dt(dtv)
        except ValueError:
            pass

    if isinstance(value, datetime.datetime):
        return ensure_pendulum_dt(value)
    elif isinstance(value, datetime.date):
        return DateTime(value.year, value.month, value.day)
    elif isinstance(value, (int, float, str)):
        result = _datetime_from_ts_or_iso(value)
        if isinstance(result, datetime.time):
            raise ValueError(f"Cannot coerce `{value}` to `pendulum.DateTime` object.")
        if isinstance(result, pendulum.DateTime):
            return result
        # naive datetime from date
        return DateTime(result.year, result.month, result.day)
    raise TypeError(f"Cannot coerce `{value}` to `pendulum.DateTime` object.")


def normalize_timezone(dt: pendulum.DateTime, timezone: bool) -> pendulum.DateTime:
    """Normalizes timezone in a pendulum instance according to dlt convention:
    * naive datetimes represent UTC (system timezone is ignored) time zone
    * tz-aware datetimes are always UTC

    Following conversions will be made:
    * when `timezone` is false: tz-aware tz is converted into UTC tz and then naive
    * when `timezone` is true: naive and aware datetimes are converted to UTC

    """
    if timezone:
        # adds UTC to naive timezones (disregard system timezone), shifts tz aware timezones
        return dt.in_tz(UTC)
    elif dt.tzinfo is not None:
        # if tz-aware then shift to UTC first, naive() just strips tz
        return dt.in_tz(tz=UTC).naive()
    return dt


def datetime_obj_to_str(
    datatime: Union[datetime.datetime, datetime.date], datetime_format: str
) -> str:
    if sys.version_info < (3, 12, 0) and "%:z" in datetime_format:
        modified_format = datetime_format.replace("%:z", "%z")
        datetime_str = datatime.strftime(modified_format)

        timezone_part = datetime_str[-5:] if len(datetime_str) >= 5 else ""
        if timezone_part.startswith(("-", "+")):
            return f"{datetime_str[:-5]}{timezone_part[:3]}:{timezone_part[3:]}"

        raise ValueError(f"Invalid timezone format in datetime string: `{datetime_str}`")

    return datatime.strftime(datetime_format)


def ensure_pendulum_time(value: Union[str, int, float, datetime.time, timedelta]) -> pendulum.Time:
    """Coerce a time-like value to a `pendulum.Time` object using timezone=False semantics.

    this follows normalize_timezone(..., timezone=False): tz-aware inputs are converted to UTC
    and then made naive; naive values are treated as UTC and kept naive.

    Args:
        value: Time value to coerce. Supported types:
            - pendulum.Time or datetime.time
            - ISO time string (e.g. "12:34:56", "12:34:56+02:00")
            - timedelta representing seconds since midnight

    Returns:
        A naive pendulum.Time object that represents UTC time-of-day.
    """

    def _normalize_aware_time(t: datetime.time) -> pendulum.Time:
        # fast path: if naive, do not normalize
        if t.tzinfo is None:
            return pendulum.time(t.hour, t.minute, t.second, t.microsecond)
        # build a dummy date to apply timezone normalization uniformly
        fold = getattr(t, "fold", 0)
        dt = create_dt(
            1970, 1, 1, t.hour, t.minute, t.second, t.microsecond, tz=t.tzinfo, fold=fold
        )
        ndt = normalize_timezone(dt, timezone=False)
        return pendulum.time(ndt.hour, ndt.minute, ndt.second, ndt.microsecond)

    if isinstance(value, datetime.time):
        # handles both python datetime.time and pendulum.Time
        return _normalize_aware_time(value)

    if isinstance(value, str):
        parsed = parse_iso_like_datetime(value)
        if isinstance(parsed, pendulum.Time):
            return _normalize_aware_time(parsed)
        # only ISO time strings are accepted
        raise ValueError(f"Invalid ISO time string: `{value}`")

    if isinstance(value, timedelta):
        # assume timedelta is seconds passed since midnight (eg. mysqlclient returns that)
        return pendulum.time(
            value.seconds // 3600,
            (value.seconds // 60) % 60,
            value.seconds % 60,
            value.microseconds,
        )
    raise TypeError(f"Cannot coerce `{value}` to `pendulum.Time` object.")


def detect_datetime_format(value: str) -> Optional[str]:
    format_patterns = {
        # Full datetime with 'Z' (UTC) or timezone offset
        re.compile(r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z$"): "%Y-%m-%dT%H:%M:%SZ",  # UTC 'Z'
        re.compile(
            r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d+Z$"
        ): "%Y-%m-%dT%H:%M:%S.%fZ",  # UTC with fractional seconds
        re.compile(
            r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\+\d{2}:\d{2}$"
        ): "%Y-%m-%dT%H:%M:%S%:z",  # Positive timezone offset
        re.compile(
            r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\+\d{4}$"
        ): "%Y-%m-%dT%H:%M:%S%z",  # Positive timezone without colon
        # Full datetime with fractional seconds and positive timezone offset
        re.compile(
            r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d+\+\d{2}:\d{2}$"
        ): "%Y-%m-%dT%H:%M:%S.%f%:z",
        re.compile(
            r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d+\+\d{4}$"
        ): "%Y-%m-%dT%H:%M:%S.%f%z",  # Positive timezone without colon
        re.compile(
            r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}-\d{2}:\d{2}$"
        ): "%Y-%m-%dT%H:%M:%S%:z",  # Negative timezone offset
        re.compile(
            r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}-\d{4}$"
        ): "%Y-%m-%dT%H:%M:%S%z",  # Negative timezone without colon
        # Full datetime with fractional seconds and negative timezone offset
        re.compile(
            r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d+-\d{2}:\d{2}$"
        ): "%Y-%m-%dT%H:%M:%S.%f%:z",
        re.compile(
            r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d+-\d{4}$"
        ): "%Y-%m-%dT%H:%M:%S.%f%z",  # Negative Timezone without colon
        # Datetime without timezone
        re.compile(r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}$"): "%Y-%m-%dT%H:%M:%S",  # No timezone
        re.compile(r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}$"): "%Y-%m-%dT%H:%M",  # Minute precision
        re.compile(r"^\d{4}-\d{2}-\d{2}T\d{2}$"): "%Y-%m-%dT%H",  # Hour precision
        # Date-only formats
        re.compile(r"^\d{4}-\d{2}-\d{2}$"): "%Y-%m-%d",  # Date only
        re.compile(r"^\d{4}-\d{2}$"): "%Y-%m",  # Year and month
        re.compile(r"^\d{4}$"): "%Y",  # Year only
        # Week-based date formats
        re.compile(r"^\d{4}-W\d{2}$"): "%Y-W%W",  # Week-based date
        re.compile(r"^\d{4}-W\d{2}-\d{1}$"): "%Y-W%W-%u",  # Week-based date with day
        # Ordinal date formats (day of year)
        re.compile(r"^\d{4}-\d{3}$"): "%Y-%j",  # Ordinal date
        # Compact formats (no dashes)
        re.compile(r"^\d{8}$"): "%Y%m%d",  # Compact date format
        re.compile(r"^\d{6}$"): "%Y%m",  # Compact year and month format
    }

    # Match against each compiled regular expression
    for pattern, format_str in format_patterns.items():
        if pattern.match(value):
            return format_str

    # Return None if no pattern matches
    return None


def to_py_datetime(value: datetime.datetime) -> datetime.datetime:
    """Convert a pendulum.DateTime to a py datetime object.

    Args:
        value: The value to convert. Can be a pendulum.DateTime or datetime.

    Returns:
        A py datetime object
    """
    if isinstance(value, pendulum.DateTime):
        return datetime.datetime(
            value.year,
            value.month,
            value.day,
            value.hour,
            value.minute,
            value.second,
            value.microsecond,
            value.tzinfo,
        )
    return value


def to_py_date(value: datetime.date) -> datetime.date:
    """Convert a pendulum.Date to a py date object.

    Args:
        value: The value to convert. Can be a pendulum.Date or date.

    Returns:
        A py date object
    """
    if isinstance(value, pendulum.Date):
        return datetime.date(value.year, value.month, value.day)
    return value


def datetime_to_timestamp(moment: Union[datetime.datetime, pendulum.DateTime]) -> int:
    return int(moment.timestamp())


def datetime_to_timestamp_ms(moment: Union[datetime.datetime, pendulum.DateTime]) -> int:
    return int(moment.timestamp() * 1000)


def _datetime_from_ts_or_iso(
    value: Union[int, float, str]
) -> Union[pendulum.DateTime, pendulum.Date, pendulum.Time]:
    if isinstance(value, (int, float)):
        return pendulum.from_timestamp(value)

    try:
        return parse_iso_like_datetime(value)
    except ValueError as outer_ex:
        try:
            value = float(value)
        except ValueError:
            raise outer_ex from None
        return pendulum.from_timestamp(value)


@overload
def to_seconds(td: None) -> None:
    pass


@overload
def to_seconds(td: TimedeltaSeconds) -> float:
    pass


def to_seconds(td: Optional[TimedeltaSeconds]) -> Optional[float]:
    if isinstance(td, timedelta):
        return td.total_seconds()
    return td


TTimeWithPrecision = TypeVar("TTimeWithPrecision", bound=Union[pendulum.DateTime, pendulum.Time])


def reduce_pendulum_datetime_precision(
    value: TTimeWithPrecision, precision: int
) -> TTimeWithPrecision:
    if precision >= 6:
        return value
    return value.replace(microsecond=value.microsecond // 10 ** (6 - precision) * 10 ** (6 - precision))  # type: ignore


def get_precision_from_datetime_unit(unit: str) -> int:
    """Convert PyArrow datetime unit to numeric precision.

    Args:
        unit: PyArrow datetime unit ("s", "ms", "us", "ns")

    Returns:
        Numeric precision (0, 3, 6, or 9)
    """
    if unit == "s":
        return 0
    elif unit == "ms":
        return 3
    elif unit == "us":
        return 6
    else:  # "ns" or any other unit defaults to nanosecond precision
        return 9



"""The Requests Pipeline Template provides a simple starting point for a dlt pipeline with the requests library"""

# mypy: disable-error-code="no-untyped-def,arg-type"

from typing import Iterator, Any

import dlt

from dlt.sources.helpers import requests
from dlt.sources import TDataItems


YEAR = 2022
MONTH = 10
BASE_PATH = "https://api.chess.com/pub/player"


@dlt.resource(primary_key="player_id")
def players():
    """Load player profiles from the chess api."""
    for player_name in ["magnuscarlsen", "rpragchess"]:
        path = f"{BASE_PATH}/{player_name}"
        response = requests.get(path)
        response.raise_for_status()
        yield response.json()


# this resource takes data from players and returns games for the configured
@dlt.transformer(data_from=players, write_disposition="append")
def players_games(player: Any) -> Iterator[TDataItems]:
    """Load all games for each player in october 2022"""
    player_name = player["username"]
    path = f"{BASE_PATH}/{player_name}/games/{YEAR:04d}/{MONTH:02d}"
    response = requests.get(path)
    response.raise_for_status()
    yield response.json()["games"]


@dlt.source(name="chess")
def chess():
    """A source function groups all resources into one schema."""
    return players(), players_games()


def load_chess_data() -> None:
    # specify the pipeline name, destination and dataset name when configuring pipeline,
    # otherwise the defaults will be used that are derived from the current script name
    p = dlt.pipeline(
        pipeline_name="chess",
        destination="duckdb",
        dataset_name="chess_data",
    )

    load_info = p.run(chess())

    # pretty print the information on data that was loaded
    print(load_info)  # noqa: T201


if __name__ == "__main__":
    load_chess_data()



from abc import ABC, abstractmethod
from typing import (
    Any,
    Callable,
    Iterator,
    Iterable,
    Literal,
    Optional,
    Protocol,
    Union,
    Awaitable,
    TYPE_CHECKING,
    NamedTuple,
)
from concurrent.futures import Future

from dlt.common.typing import (
    TAny,
    TDataItem,
    TDataItems,
    TTableHintTemplate,
    TFunHintTemplate,
    TDynHintType,
)

TDecompositionStrategy = Literal["none", "scc"]
TDeferredDataItems = Callable[[], TDataItems]
TAwaitableDataItems = Awaitable[TDataItems]
TPipedDataItems = Union[TDataItems, TDeferredDataItems, TAwaitableDataItems]


if TYPE_CHECKING:
    TItemFuture = Future[TPipedDataItems]
else:
    TItemFuture = Future


class PipeItem(NamedTuple):
    item: TDataItems
    step: int
    pipe: "SupportsPipe"
    meta: Any


class ResolvablePipeItem(NamedTuple):
    # mypy unable to handle recursive types, ResolvablePipeItem should take itself in "item"
    item: Union[TPipedDataItems, Iterator[TPipedDataItems]]
    step: int
    pipe: "SupportsPipe"
    meta: Any


class FuturePipeItem(NamedTuple):
    item: TItemFuture
    step: int
    pipe: "SupportsPipe"
    meta: Any


class SourcePipeItem(NamedTuple):
    item: Union[Iterator[TPipedDataItems], Iterator[ResolvablePipeItem]]
    step: int
    pipe: "SupportsPipe"
    meta: Any


# pipeline step may be iterator of data items or mapping function that returns data item or another iterator
TPipeStep = Union[
    Iterable[TPipedDataItems],
    Iterator[TPipedDataItems],
    # Callable with meta
    Callable[[TDataItems, Optional[Any]], TPipedDataItems],
    Callable[[TDataItems, Optional[Any]], Iterator[TPipedDataItems]],
    Callable[[TDataItems, Optional[Any]], Iterator[ResolvablePipeItem]],
    # Callable without meta
    Callable[[TDataItems], TPipedDataItems],
    Callable[[TDataItems], Iterator[TPipedDataItems]],
    Callable[[TDataItems], Iterator[ResolvablePipeItem]],
]


class DataItemWithMeta:
    __slots__ = ("meta", "data")

    def __init__(self, meta: Any, data: TDataItems) -> None:
        self.meta = meta
        self.data = data


class TableNameMeta:
    __slots__ = ("table_name",)

    def __init__(self, table_name: str) -> None:
        self.table_name = table_name


class SupportsPipe(Protocol):
    """A protocol with the core Pipe properties and operations"""

    name: str
    """Pipe name which is inherited by a resource"""
    parent: "SupportsPipe"
    """A parent of the current pipe"""
    instance_id: str
    """Unique id of the pipe assigned when instance is created and preserved when cloning"""

    @property
    def gen(self) -> TPipeStep:
        """A data generating step"""
        ...

    def replace_gen(self, gen: TPipeStep) -> None:
        """Replaces data generating step. Assumes that you know what are you doing"""
        ...

    def __getitem__(self, i: int) -> TPipeStep:
        """Get pipe step at index"""
        ...

    def __len__(self) -> int:
        """Length of a pipe"""
        ...

    @property
    def has_parent(self) -> bool:
        """Checks if pipe is connected to parent pipe from which it takes data items. Connected pipes are created from transformer resources"""
        ...

    def close(self) -> None:
        """Closes pipe generator"""
        ...



"""The Arrow Pipeline Template will show how to load and transform arrow tables."""

# mypy: disable-error-code="no-untyped-def,arg-type"

import dlt
import time
import pyarrow as pa


def create_example_arrow_table() -> pa.Table:
    return pa.Table.from_pylist([{"name": "tom", "age": 25}, {"name": "angela", "age": 23}])


@dlt.resource(write_disposition="append", name="people")
def resource():
    """One resource function will materialize as a table in the destination, wie yield example data here"""
    yield create_example_arrow_table()


def add_updated_at(item: pa.Table):
    """Map function to add an updated at column to your incoming data."""
    column_count = len(item.columns)
    # you will receive and return and arrow table
    return item.set_column(column_count, "updated_at", [[time.time()] * item.num_rows])


# apply transformer to resource
resource.add_map(add_updated_at)


@dlt.source
def source():
    """A source function groups all resources into one schema."""
    # return resources
    return resource()


def load_arrow_tables() -> None:
    # specify the pipeline name, destination and dataset name when configuring pipeline,
    # otherwise the defaults will be used that are derived from the current script name
    pipeline = dlt.pipeline(
        pipeline_name="arrow",
        destination="duckdb",
        dataset_name="arrow_data",
    )

    data = list(source().people)

    # print the data yielded from resource without loading it
    print(data)  # noqa: T201

    # run the pipeline with your parameters
    load_info = pipeline.run(source())

    # pretty print the information on data that was loaded
    print(load_info)  # noqa: T201


if __name__ == "__main__":
    load_arrow_tables()



from dlt.common.exceptions import DltException


class NormalizersException(DltException):
    pass


class UnknownNamingModule(ImportError, NormalizersException):
    def __init__(self, naming_module: str) -> None:
        self.naming_module = naming_module
        if "." in naming_module:
            msg = f"Naming module `{naming_module}` could not be found and imported"
        else:
            msg = (
                f"Naming module `{naming_module} `is not one of the standard dlt naming conventions"
                " and could not be locally imported"
            )
        super().__init__(msg)


class NamingTypeNotFound(ImportError, NormalizersException):
    def __init__(self, naming_module: str, naming_class: str) -> None:
        self.naming_module = naming_module
        self.naming_class = naming_class
        msg = f"In naming module `{naming_module}` type `{naming_class}` does not exist"
        super().__init__(msg)


class InvalidNamingType(NormalizersException):
    def __init__(self, naming_module: str, naming_class: str) -> None:
        self.naming_module = naming_module
        self.naming_class = naming_class
        msg = (
            f"In naming module `{naming_module}` the class `{naming_class} `is not a"
            " `NamingConvention`"
        )
        super().__init__(msg)



from typing import List, Optional

import sqlglot

from dlt.common.libs.sqlglot import (
    TSqlGlotDialect,
    SqlModel,
    build_outer_select_statement,
    reorder_or_adjust_outer_select,
    uuid_expr_for_dialect,
    validate_no_star_select,
)
from dlt.common.normalizers.json.helpers import get_root_row_id_type
from dlt.common.schema.typing import C_DLT_ID, C_DLT_LOAD_ID
from dlt.common.schema.utils import (
    dlt_id_column,
    dlt_load_id_column,
    normalize_table_identifiers,
)
from dlt.common.schema import TSchemaUpdate
from dlt.common.utils import read_dialect_and_sql

from dlt.normalize.exceptions import NormalizeException
from dlt.normalize.items_normalizers.base import ItemsNormalizer


class ModelItemsNormalizer(ItemsNormalizer):
    def _normalize_casefold(self, identifier: str) -> str:
        # use normalize_path() to preserve __ separators in already-normalized names
        normalized = self.schema.naming.normalize_path(identifier)
        return self.config.destination_capabilities.casefold_identifier(normalized)

    def _adjust_outer_select_with_dlt_columns(
        self,
        sql_dialect: TSqlGlotDialect,
        outer_parsed_select: sqlglot.exp.Select,
        root_table_name: str,
    ) -> Optional[TSchemaUpdate]:
        """
        Adds or replaces dlt-specific columns (`_dlt_id`, `_dlt_load_id`) in the SELECT statement.

        Args:
            outer_parsed_select (sqlglot.exp.Select): The parsed outer SELECT statement.
            root_table_name (str): The name of the root table being normalized.

        Returns:
            Optional[TSchemaUpdate]: Schema updates for the added or replaced columns, or None if no updates were made.
        """
        model_config = self.config.model_normalizer
        if not (model_config.add_dlt_load_id or model_config.add_dlt_id):
            return None

        schema_update: TSchemaUpdate = {}
        schema = self.schema

        existing = {
            select.alias.lower(): idx for idx, select in enumerate(outer_parsed_select.selects)
        }

        NORM_C_DLT_LOAD_ID = self._normalize_casefold(C_DLT_LOAD_ID)
        NORM_C_DLT_ID = self._normalize_casefold(C_DLT_ID)

        # 1. Handle _dlt_load_id addition
        if model_config.add_dlt_load_id:
            # Build aliased expression
            dlt_load_id_expr = sqlglot.exp.Alias(
                this=sqlglot.exp.Literal.string(self.load_id),
                alias=sqlglot.to_identifier(NORM_C_DLT_LOAD_ID, quoted=True),
            )
            # Replace in-place if already present, otherwise append and update schema
            idx = existing.get(C_DLT_LOAD_ID)
            if idx is not None:
                outer_parsed_select.selects[idx] = dlt_load_id_expr
            else:
                outer_parsed_select.selects.append(dlt_load_id_expr)
                partial_table = normalize_table_identifiers(
                    {
                        "name": root_table_name,
                        "columns": {C_DLT_LOAD_ID: dlt_load_id_column()},
                    },
                    schema.naming,
                )
                schema.update_table(partial_table)
                table_updates = schema_update.setdefault(root_table_name, [])
                table_updates.append(partial_table)

        # 2. Handle _dlt_id addition
        if model_config.add_dlt_id:
            idx = existing.get(C_DLT_ID)
            # Do nothing if already present,
            # otherwise append and update schema only if possible
            if idx is None:
                row_id_type = get_root_row_id_type(schema, root_table_name)
                if row_id_type != "random":
                    raise NormalizeException(
                        "`add_dlt_id` was enabled in the model normalizer config, "
                        "but this is only supported when the row id type equals 'random'`. "
                        f"Received row id type: '{row_id_type}'."
                    )
                dlt_id_expr = sqlglot.exp.Alias(
                    this=uuid_expr_for_dialect(sql_dialect, self.load_id),
                    alias=sqlglot.to_identifier(NORM_C_DLT_ID, quoted=True),
                )
                outer_parsed_select.selects.append(dlt_id_expr)
                partial_table = normalize_table_identifiers(
                    {
                        "name": root_table_name,
                        "columns": {C_DLT_ID: dlt_id_column()},
                    },
                    schema.naming,
                )
                schema.update_table(partial_table)
                table_updates = schema_update.setdefault(root_table_name, [])
                table_updates.append(partial_table)

        return schema_update or None

    def __call__(self, extracted_items_file: str, root_table_name: str) -> List[TSchemaUpdate]:
        self._maybe_cancel()
        with self.normalize_storage.extracted_packages.storage.open_file(
            extracted_items_file, "r"
        ) as f:
            sql_dialect, select_statement = read_dialect_and_sql(
                file_obj=f,
                fallback_dialect=self.config.destination_capabilities.sqlglot_dialect,  # caps are available at this point
            )

        # TODO the dialect here should be the "query dialect"; i.e., the transpilation input
        parsed_select = sqlglot.parse_one(select_statement, read=sql_dialect)

        # The query is ensured to be a select statement upstream,
        # but we double check here
        if not isinstance(parsed_select, sqlglot.exp.Select):
            raise ValueError("Only SELECT statements should be used as `SqlModel` queries.")

        validate_no_star_select(parsed_select, sql_dialect)

        outer_parsed_select, needs_reordering = build_outer_select_statement(
            sql_dialect,
            parsed_select,
            self.schema.get_table_columns(root_table_name),
            self._normalize_casefold,
        )

        schema_updates = []
        dlt_col_update = self._adjust_outer_select_with_dlt_columns(
            sql_dialect, outer_parsed_select, root_table_name
        )

        if dlt_col_update:
            schema_updates.append(dlt_col_update)

        if needs_reordering:
            reorder_or_adjust_outer_select(
                outer_parsed_select,
                self.schema.get_table_columns(root_table_name),
                self._normalize_casefold,
                self.schema.name,
                root_table_name,
            )

        # TODO the dialect here should be the "destination dialect"; i.e., the transpilation output
        normalized_query = outer_parsed_select.sql(dialect=sql_dialect)
        self.item_storage.write_data_item(
            self.load_id,
            self.schema.name,
            root_table_name,
            SqlModel.from_query_string(normalized_query, sql_dialect),
            {},
        )

        return schema_updates



import re
import os
from typing import Any, Dict, List, Optional, Sequence, Set, Tuple

from dlt.common import logger
from dlt.common.pendulum import pendulum
from dlt.common.storages.load_package import ParsedLoadJobFileName
from dlt.common.time import (
    ensure_pendulum_datetime_utc,
    datetime_to_timestamp,
    datetime_to_timestamp_ms,
)
from dlt.destinations.exceptions import (
    CantExtractTablePrefix,
    InvalidFilesystemLayout,
    InvalidPlaceholderCallback,
)
from dlt.destinations.impl.filesystem.typing import TCurrentDateTime


# For formatting options please see
# https://github.com/sdispater/pendulum/blob/master/docs/docs/string_formatting.md
DATETIME_PLACEHOLDERS = {
    # Years
    "YYYY",  # 2024, 2025, 2026
    "YY",  # 00, 01, 02 ... 12, 13
    "Y",  # 2024, 2025, 2026
    # Months
    "MMMM",  # January, February, March
    "MMM",  # Jan, Feb, Mar
    "MM",  # 01-12
    "M",  # 1-12
    # Days
    "DD",  # 01-31
    "D",  # 1-31
    # Hours
    "HH",  # 00-23
    "H",  # 0-23
    # Minutes
    "mm",  # 00-59
    "m",  # 0-59
    # Days of week
    "dddd",  # Monday, Tuesday, Wednesday
    "ddd",  # Mon, Tue, Wed
    "dd",  # Mo, Tu, We
    "d",  # 0-6
    # Seconds
    "ss",  # 01-59
    "s",  # 0-59
    # Fractional seconds
    "SSSS",  # 000[0..] 001[0..] ... 998[0..] 999[0..]
    "SSS",  # 000 001 ... 998 999
    "SS",  # 00, 01, 02 ... 98, 99
    "S",  # 0 1 ... 8 9
    # Quarters of the year
    "Q",  # 1, 2, 3, 4
}

STANDARD_PLACEHOLDERS = DATETIME_PLACEHOLDERS.union(
    {
        "schema_name",
        "table_name",
        "load_id",
        "file_id",
        "ext",
        "curr_date",
        "timestamp",
        "timestamp",
        "timestamp_ms",
        "load_package_timestamp",
        "load_package_timestamp_ms",
    }
)


SUPPORTED_TABLE_NAME_PREFIX_PLACEHOLDERS = ("schema_name",)


def normalize_path_sep(pathlib: Any, path: str) -> str:
    """Normalizes path in `path` separator to one used by `pathlib`"""
    if pathlib.sep == "/":
        return path.replace("\\", "/")
    if pathlib.sep == "\\":
        return path.replace("/", "\\")
    return path


def get_placeholders(layout: str) -> List[str]:
    return re.findall(r"\{(.*?)\}", layout)


def get_unused_placeholders(
    layout_placeholders: Sequence[str],
    extra_placeholders: Sequence[str],
) -> Sequence[str]:
    unused_placeholders = [p for p in extra_placeholders if p not in layout_placeholders]
    return unused_placeholders


def prepare_datetime_params(
    current_datetime: Optional[pendulum.DateTime] = None,
    load_package_timestamp: Optional[pendulum.DateTime] = None,
) -> Dict[str, str]:
    params: Dict[str, str] = {}
    current_timestamp: pendulum.DateTime = None
    if load_package_timestamp:
        current_timestamp = ensure_pendulum_datetime_utc(load_package_timestamp)
        params["load_package_timestamp"] = str(datetime_to_timestamp(current_timestamp))
        params["load_package_timestamp_ms"] = str(datetime_to_timestamp_ms(current_timestamp))

    if not current_datetime:
        if current_timestamp:
            logger.info("current_datetime is not set, using timestamp from load package")
            current_datetime = current_timestamp
        else:
            logger.info("current_datetime is not set, using pendulum.now()")
            current_datetime = pendulum.now()

    params["timestamp"] = str(datetime_to_timestamp(current_datetime))
    params["timestamp_ms"] = str(datetime_to_timestamp_ms(current_datetime))
    params["curr_date"] = str(current_datetime.date())

    for format_string in DATETIME_PLACEHOLDERS:
        params[format_string] = current_datetime.format(format_string).lower()

    return params


def prepare_params(
    extra_placeholders: Optional[Dict[str, Any]] = None,
    job_info: Optional[ParsedLoadJobFileName] = None,
    schema_name: Optional[str] = None,
    load_id: Optional[str] = None,
) -> Dict[str, Any]:
    params: Dict[str, Any] = {}
    table_name = None
    file_id = None
    ext = None
    if job_info:
        table_name = job_info.table_name
        file_id = job_info.file_id
        ext = job_info.full_extension()
        params.update(
            {
                "table_name": table_name,
                "file_id": file_id,
                "ext": ext,
            }
        )

    if schema_name:
        params["schema_name"] = schema_name

    if load_id:
        params["load_id"] = load_id

    # Resolve extra placeholders
    if extra_placeholders:
        for key, value in extra_placeholders.items():
            if callable(value):
                try:
                    params[key] = value(
                        schema_name,
                        table_name,
                        load_id,
                        file_id,
                        ext,
                    )
                except TypeError as exc:
                    raise InvalidPlaceholderCallback(key) from exc
            else:
                params[key] = value

    return params


def check_layout(
    layout: str,
    extra_placeholders: Optional[Dict[str, Any]] = None,
    standard_placeholders: Optional[Set[str]] = STANDARD_PLACEHOLDERS,
) -> Tuple[List[str], List[str]]:
    """Returns a tuple with all valid placeholders and the list of layout placeholders

    Raises: InvalidFilesystemLayout

    Returns: a pair of lists of valid placeholders and layout placeholders
    """
    placeholders = get_placeholders(layout)
    # Build out the list of placeholder names
    # which we will use to validate placeholders
    # in a given config.layout template
    all_placeholders = standard_placeholders.copy()
    if extra_placeholders:
        for placeholder, _ in extra_placeholders.items():
            all_placeholders.add(placeholder)

    # now collect all unknown placeholders from config.layout template
    invalid_placeholders = [p for p in placeholders if p not in all_placeholders]
    extra_placeholder_keys = list((extra_placeholders or {}).keys())
    unused_placeholders = get_unused_placeholders(placeholders, extra_placeholder_keys)
    if invalid_placeholders:
        raise InvalidFilesystemLayout(
            layout,
            all_placeholders,  # type: ignore[arg-type]
            extra_placeholder_keys,
            invalid_placeholders,
            unused_placeholders,
        )

    return list(all_placeholders), placeholders


def create_path(
    layout: str,
    file_name: str,
    schema_name: str,
    load_id: str,
    load_package_timestamp: Optional[pendulum.DateTime] = None,
    current_datetime: Optional[TCurrentDateTime] = None,
    extra_placeholders: Optional[Dict[str, Any]] = None,
) -> str:
    """create a filepath from the layout and our default params"""
    if callable(current_datetime):
        current_datetime = current_datetime()
        if not isinstance(current_datetime, pendulum.DateTime):
            raise RuntimeError("`current_datetime` is not an instance of `pendulum.DateTime`")

    job_info = ParsedLoadJobFileName.parse(file_name)
    params = prepare_params(
        extra_placeholders=extra_placeholders,
        job_info=job_info,
        schema_name=schema_name,
        load_id=load_id,
    )

    datetime_params = prepare_datetime_params(current_datetime, load_package_timestamp)
    params.update(datetime_params)

    placeholders, _ = check_layout(layout, params)
    path = layout.format(**params)

    # if extension is not defined, we append it at the end
    if "ext" not in placeholders:
        path = job_info.full_extension()

    return path


def get_table_prefix_layout(
    layout: str,
    supported_prefix_placeholders: Sequence[str] = SUPPORTED_TABLE_NAME_PREFIX_PLACEHOLDERS,
    table_needs_own_folder: bool = False,
) -> str:
    """get layout fragment that defines positions of the table, cutting other placeholders
    allowed `supported_prefix_placeholders` that may appear before table.
    """
    placeholders = get_placeholders(layout)
    # fail if table name is not defined
    if "table_name" not in placeholders:
        raise CantExtractTablePrefix(layout, "{table_name} placeholder not found. ")

    table_name_index = placeholders.index("table_name")

    # fail if any other prefix is defined before table_name
    if [p for p in placeholders[:table_name_index] if p not in supported_prefix_placeholders]:
        if len(supported_prefix_placeholders) == 0:
            details = (
                "No other placeholders are allowed before {table_name} but you have %s present. "
                % placeholders[:table_name_index]
            )
        else:
            details = "Only %s are allowed before {table_name} but you have %s present. " % (
                supported_prefix_placeholders,
                placeholders[:table_name_index],
            )
        raise CantExtractTablePrefix(layout, details)

    # we include the char after the table_name here, this should be a separator not a new placeholder
    # this is to prevent selecting tables that have the same starting name -> {table_name}/
    prefix = layout[: layout.index("{table_name}") + 13]
    if prefix[-1] == "{":
        raise CantExtractTablePrefix(layout, "A separator is required after a {table_name}. ")
    if prefix[-1] != "/" and table_needs_own_folder:
        raise CantExtractTablePrefix(
            layout, "Table requires it's own folder, please add a '/' after your {table_name}. "
        )

    return prefix


def get_file_format_and_compression(file_path: str) -> Tuple[str, bool]:
    """Return file format and whether an explicit .gz compression extension is present.
    Primarily used in cases that ParsedLoadJobFileName.parse() can't handle, that is,
    staging storage files that have a configurabe layout."""
    root, ext = os.path.splitext(file_path)

    file_format: str = None
    compression_ext: bool = False

    if ext == ".gz":
        compression_ext = True
        _, ext = os.path.splitext(root)
        file_format = ext[1:]  # remove the dot
    else:
        file_format = ext[1:]  # remove the dot

    return file_format, compression_ext



import dataclasses
from typing import ClassVar, Final, Optional, Any, Dict, List, List, Dict, cast, Callable
from urllib.parse import urlparse

from dlt.common import logger
from dlt.common.typing import TSecretStrValue
from dlt.common.configuration.specs.base_configuration import CredentialsConfiguration, configspec
from dlt.common.destination.client import DestinationClientDwhWithStagingConfiguration
from dlt.common.configuration.exceptions import ConfigurationValueError
from dlt.common.utils import digest128

DATABRICKS_APPLICATION_ID = "dltHub_dlt"


@configspec
class DatabricksCredentials(CredentialsConfiguration):
    catalog: str = None
    server_hostname: Optional[str] = None
    http_path: Optional[str] = None
    access_token: Optional[TSecretStrValue] = None
    client_id: Optional[TSecretStrValue] = None
    client_secret: Optional[TSecretStrValue] = None
    http_headers: Optional[Dict[str, str]] = None
    session_configuration: Optional[Dict[str, Any]] = None
    """Dict of session parameters that will be passed to `databricks.sql.connect`"""
    connection_parameters: Optional[Dict[str, Any]] = None
    """Additional keyword arguments that are passed to `databricks.sql.connect`"""
    socket_timeout: Optional[int] = 180
    user_agent_entry: Optional[str] = DATABRICKS_APPLICATION_ID

    __config_gen_annotations__: ClassVar[List[str]] = [
        "server_hostname",
        "http_path",
        "catalog",
        "client_id",
        "client_secret",
        "access_token",
    ]

    def on_resolved(self) -> None:
        try:
            from databricks.sdk import WorkspaceClient

            w = WorkspaceClient()
        except Exception:
            w = None

        if not ((self.client_id and self.client_secret) or self.access_token):
            if w is not None:
                try:
                    # attempt context authentication via notebook token
                    self.access_token = (
                        w.dbutils.notebook.entry_point.getDbutils()
                        .notebook()
                        .getContext()
                        .apiToken()
                        .getOrElse(None)
                    )
                except Exception:
                    self.access_token = None

                try:
                    self.access_token = w.config.authenticate
                    logger.info(f"Will attempt to use default auth of type {w.config.auth_type}")
                except Exception:
                    pass

            if not self.access_token:
                raise ConfigurationValueError(
                    "Authentication failed: No valid authentication method detected. "
                    "Provide either `client_id` and `client_secret` for OAuth authentication, "
                    "or `access_token` for token-based authentication."
                )

        if not self.server_hostname or not self.http_path:
            if w is not None:
                # try to get server_hostname from workspace URL
                if not self.server_hostname:
                    try:
                        workspace_url = w.config.host
                        if workspace_url:
                            self.server_hostname = urlparse(workspace_url).hostname
                            logger.info(
                                f"Using server_hostname from workspace: {self.server_hostname}"
                            )
                    except Exception:
                        pass

                # try to get http_path from notebook cluster context
                if not self.http_path:
                    try:
                        notebook_context = (
                            w.dbutils.notebook.entry_point.getDbutils().notebook().getContext()
                        )
                        cluster_id = notebook_context.clusterId().get()
                        workspace_id = notebook_context.workspaceId().get()
                        if cluster_id and workspace_id:
                            self.http_path = f"/sql/protocolv1/o/{workspace_id}/{cluster_id}"
                            logger.info(f"Using http_path from cluster context: {self.http_path}")
                    except Exception:
                        pass

                # fallback: try to fetch warehouse details
                if not self.server_hostname or not self.http_path:
                    try:
                        if w.config.warehouse_id:
                            warehouse = w.warehouses.get(w.config.warehouse_id)
                        else:
                            # for some reason list of warehouses has different type
                            # than a single one
                            warehouse = list(w.warehouses.list())[0]
                        logger.info(
                            "Will attempt to use warehouse"
                            f" {warehouse.id} to get sql connection params"
                        )
                        self.server_hostname = (
                            self.server_hostname or warehouse.odbc_params.hostname
                        )
                        self.http_path = self.http_path or warehouse.odbc_params.path
                    except Exception:
                        pass

        for param in ("catalog", "server_hostname", "http_path"):
            if not getattr(self, param):
                raise ConfigurationValueError(
                    f"Configuration error: Missing required parameter `{param}`. "
                    "Please provide it in the configuration."
                )

    def _get_oauth_credentials(self) -> Optional[Callable[[], Dict[str, str]]]:
        from databricks.sdk.core import Config, oauth_service_principal

        config = Config(
            host=f"https://{self.server_hostname}",
            client_id=self.client_id,
            client_secret=self.client_secret,
        )
        return cast(Callable[[], Dict[str, str]], oauth_service_principal(config))

    def to_connector_params(self) -> Dict[str, Any]:
        conn_params = dict(
            catalog=self.catalog,
            server_hostname=self.server_hostname,
            http_path=self.http_path,
            access_token=self.access_token,
            session_configuration=self.session_configuration or {},
            _socket_timeout=self.socket_timeout,
            **(self.connection_parameters or {}),
        )

        if self.user_agent_entry:
            conn_params["_user_agent_entry"] = (
                conn_params.get("_user_agent_entry") or self.user_agent_entry
            )

        if self.client_id and self.client_secret:
            conn_params["credentials_provider"] = self._get_oauth_credentials
        elif callable(self.access_token):
            # this is w.config.authenticator
            conn_params["credentials_provider"] = lambda: self.access_token
        else:
            # this is access token
            conn_params["access_token"] = self.access_token
        return conn_params

    def __str__(self) -> str:
        return f"databricks://{self.server_hostname}{self.http_path}/{self.catalog}"


@configspec
class DatabricksClientConfiguration(DestinationClientDwhWithStagingConfiguration):
    destination_type: Final[str] = dataclasses.field(default="databricks", init=False, repr=False, compare=False)  # type: ignore[misc]
    credentials: DatabricksCredentials = None
    staging_credentials_name: Optional[str] = None
    "If set, credentials with given name will be used in copy command"
    is_staging_external_location: bool = False
    """If true, the temporary credentials are not propagated to the COPY command"""
    staging_volume_name: Optional[str] = None
    """Name of the Databricks managed volume for temporary storage, e.g., <catalog_name>.<database_name>.<volume_name>. Defaults to '_dlt_temp_load_volume' if not set."""
    keep_staged_files: Optional[bool] = True
    """Tells if to keep the files in internal (volume) stage"""

    """Whether PRIMARY KEY or FOREIGN KEY constrains should be created"""
    create_indexes: bool = False

    def __str__(self) -> str:
        """Return displayable destination location"""
        if self.credentials:
            return str(self.credentials)
        else:
            return ""

    def fingerprint(self) -> str:
        """Returns a fingerprint of host part of a connection string"""
        if self.credentials and self.credentials.server_hostname:
            return digest128(self.credentials.server_hostname)
        return ""



import struct
from datetime import datetime, timedelta, timezone  # noqa: I251

from dlt.common.destination import DestinationCapabilitiesContext

import pyodbc

from contextlib import contextmanager
from typing import Any, AnyStr, ClassVar, Iterator, Optional, Sequence, Tuple

from dlt.destinations.exceptions import (
    DatabaseTerminalException,
    DatabaseTransientException,
    DatabaseUndefinedRelation,
)
from dlt.destinations.typing import DBApi, DBTransaction
from dlt.destinations.sql_client import (
    DBApiCursorImpl,
    SqlClientBase,
    raise_database_error,
    raise_open_connection_error,
)

from dlt.destinations.impl.mssql.configuration import MsSqlCredentials
from dlt.common.destination.dataset import DBApiCursor


def handle_datetimeoffset(dto_value: bytes) -> datetime:
    # ref: https://github.com/mkleehammer/pyodbc/issues/134#issuecomment-281739794
    tup = struct.unpack("<6hI2h", dto_value)  # e.g., (2017, 3, 16, 10, 35, 18, 500000000, -6, 0)
    return datetime(
        tup[0],
        tup[1],
        tup[2],
        tup[3],
        tup[4],
        tup[5],
        tup[6] // 1000,
        timezone(timedelta(hours=tup[7], minutes=tup[8])),
    )


class PyOdbcMsSqlClient(SqlClientBase[pyodbc.Connection], DBTransaction):
    dbapi: ClassVar[DBApi] = pyodbc

    def __init__(
        self,
        dataset_name: str,
        staging_dataset_name: str,
        credentials: MsSqlCredentials,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        super().__init__(credentials.database, dataset_name, staging_dataset_name, capabilities)
        self._conn: pyodbc.Connection = None
        self.credentials = credentials

    def open_connection(self) -> pyodbc.Connection:
        self._conn = pyodbc.connect(
            self.credentials.to_odbc_dsn(),
            timeout=self.credentials.connect_timeout,
        )
        # https://github.com/mkleehammer/pyodbc/wiki/Using-an-Output-Converter-function
        self._conn.add_output_converter(-155, handle_datetimeoffset)
        self._conn.autocommit = True
        return self._conn

    @raise_open_connection_error
    def close_connection(self) -> None:
        if self._conn:
            self._conn.close()
            self._conn = None

    @contextmanager
    def begin_transaction(self) -> Iterator[DBTransaction]:
        try:
            self._conn.autocommit = False
            yield self
            self.commit_transaction()
        except Exception:
            self.rollback_transaction()
            raise

    @raise_database_error
    def commit_transaction(self) -> None:
        self._conn.commit()
        self._conn.autocommit = True

    @raise_database_error
    def rollback_transaction(self) -> None:
        try:
            self._conn.rollback()
        except pyodbc.ProgrammingError as ex:
            if (
                ex.args[0] == "42000" and "(111214)" in ex.args[1]
            ):  # "no corresponding transaction found"
                pass  # there was nothing to rollback, we silently ignore the error
            else:
                raise
        finally:
            self._conn.autocommit = True

    @property
    def native_connection(self) -> pyodbc.Connection:
        return self._conn

    def drop_dataset(self) -> None:
        # MS Sql doesn't support DROP ... CASCADE, drop tables in the schema first
        # Drop all views
        rows = self.execute_sql(
            "SELECT table_name FROM INFORMATION_SCHEMA.VIEWS WHERE table_schema = %s",
            self.capabilities.casefold_identifier(self.dataset_name),
        )
        view_names = [row[0] for row in rows]
        self._drop_views(*view_names)
        # Drop all tables
        rows = self.execute_sql(
            "SELECT table_name FROM INFORMATION_SCHEMA.TABLES WHERE table_schema = %s",
            self.capabilities.casefold_identifier(self.dataset_name),
        )
        table_names = [row[0] for row in rows]
        self.drop_tables(*table_names)
        # Drop schema
        self._drop_schema()

    def _drop_views(self, *tables: str) -> None:
        if not tables:
            return
        statements = [
            f"DROP VIEW IF EXISTS {self.make_qualified_table_name(table)}" for table in tables
        ]
        self.execute_many(statements)

    def _drop_schema(self) -> None:
        self.execute_sql("DROP SCHEMA %s" % self.fully_qualified_dataset_name())

    def execute_sql(
        self, sql: AnyStr, *args: Any, **kwargs: Any
    ) -> Optional[Sequence[Sequence[Any]]]:
        with self.execute_query(sql, *args, **kwargs) as curr:
            if curr.description is None:
                return None
            else:
                f = curr.fetchall()
                return f

    @contextmanager
    @raise_database_error
    def execute_query(self, query: AnyStr, *args: Any, **kwargs: Any) -> Iterator[DBApiCursor]:
        assert isinstance(query, str)
        if kwargs:
            raise NotImplementedError("pyodbc does not support named parameters in queries")
        if args:
            # TODO: this is bad. See duckdb & athena also
            query = query.replace("%s", "?")
        # NOTE: do not convert it into context manager. it does not close the cursor!
        curr = self._conn.cursor()
        try:
            # unpack because empty tuple gets interpreted as a single argument
            # https://github.com/mkleehammer/pyodbc/wiki/Features-beyond-the-DB-API#passing-parameters
            curr.execute(query, *args)
            # NOTE: firsts recordset is wrapped in a cursor
            yield DBApiCursorImpl(curr)
            # clear all pending result sets
            try:
                while curr.nextset():
                    pass
            except pyodbc.Error:
                pass
        except pyodbc.Error:
            # clear all pending result sets
            try:
                while curr.nextset():
                    pass
            except pyodbc.Error:
                pass
            # immediately rollback transaction
            try:
                self._conn.rollback()
            except pyodbc.Error:
                pass
            raise
        finally:
            # clear all pending result sets
            while curr.nextset():
                pass
            # always close cursor
            curr.close()

    @classmethod
    def _make_database_exception(cls, ex: Exception) -> Exception:
        if isinstance(ex, pyodbc.ProgrammingError):
            if ex.args[0] == "42S02":
                return DatabaseUndefinedRelation(ex)
            # certain pyodbc exceptions do not have second argument
            if len(ex.args) > 1:
                if ex.args[1] == "HY000":
                    return DatabaseTransientException(ex)
                elif ex.args[0] == "42000":
                    if "(15151)" in ex.args[1]:
                        return DatabaseUndefinedRelation(ex)
                    return DatabaseTransientException(ex)
        elif isinstance(ex, pyodbc.OperationalError):
            return DatabaseTransientException(ex)
        elif isinstance(ex, pyodbc.Error):
            if ex.args[0] == "07002":  # incorrect number of arguments supplied
                return DatabaseTransientException(ex)
        return DatabaseTerminalException(ex)

    @staticmethod
    def is_dbapi_exception(ex: Exception) -> bool:
        return isinstance(ex, pyodbc.Error)

    def _limit_clause_sql(self, limit: int) -> Tuple[str, str]:
        return f"TOP ({limit})", ""



from requests import Session as BaseSession
from typing import Optional, TYPE_CHECKING, Union, Tuple, TypeVar

from dlt.sources.helpers.requests.typing import TRequestTimeout
from dlt.common.typing import TimedeltaSeconds
from dlt.common.time import to_seconds
from dlt.version import __version__


TSession = TypeVar("TSession", bound=BaseSession)


DEFAULT_TIMEOUT = 60


def _timeout_to_seconds(timeout: TRequestTimeout) -> Optional[Union[Tuple[float, float], float]]:
    return (
        (to_seconds(timeout[0]), to_seconds(timeout[1]))
        if isinstance(timeout, tuple)
        else to_seconds(timeout)
    )


class Session(BaseSession):
    """Requests session which by default adds a timeout to all requests and calls `raise_for_status()` on response

    Args:
        timeout: Timeout for requests in seconds. May be passed as `timedelta` or `float/int` number of seconds.
            May be a single value or a tuple for separate (connect, read) timeout.
        raise_for_status: Whether to raise exception on error status codes (using `response.raise_for_status()`)
    """

    def __init__(
        self,
        timeout: Optional[
            Union[TimedeltaSeconds, Tuple[TimedeltaSeconds, TimedeltaSeconds]]
        ] = DEFAULT_TIMEOUT,
        raise_for_status: bool = True,
    ) -> None:
        super().__init__()
        self.timeout = _timeout_to_seconds(timeout)
        self.raise_for_status = raise_for_status
        self.headers.update(
            {
                "User-Agent": f"dlt/{__version__}",
            }
        )

    if TYPE_CHECKING:
        request = BaseSession.request

    def request(self, *args, **kwargs):  # type: ignore[no-untyped-def,no-redef]
        kwargs.setdefault("timeout", self.timeout)
        resp = super().request(*args, **kwargs)
        if self.raise_for_status:
            resp.raise_for_status()
        return resp

    def send(self, request, **kwargs):  # type: ignore[no-untyped-def]
        kwargs.setdefault("timeout", self.timeout)
        return super().send(request, **kwargs)



import typing as t
import warnings

from dlt.common.warnings import Dlt04DeprecationWarning


def full_refresh_argument_deprecated(caller_name: str, full_refresh: t.Optional[bool]) -> None:
    """full_refresh argument is replaced with dev_mode"""
    if full_refresh is None:
        return

    warnings.warn(
        f"The `full_refresh` argument to {caller_name} is deprecated and will be removed in a"
        f" future version. Use `dev_mode={full_refresh}` instead which will have the same effect.",
        Dlt04DeprecationWarning,
        stacklevel=2,
    )



import contextlib
import dataclasses
import io
from typing import ClassVar, List

from dlt.common.configuration.exceptions import DuplicateConfigProviderException
from dlt.common.configuration.providers import (
    ConfigProvider,
    ContextProvider,
)
from dlt.common.configuration.specs import (
    GcpServiceAccountCredentials,
    BaseConfiguration,
    configspec,
    known_sections,
)


@configspec
class VaultProviderConfiguration(BaseConfiguration):
    only_secrets: bool = True
    "Only checks for secrets, ignoring all other field types"
    only_toml_fragments: bool = True
    "Only checks for toml fragments like sources or destinations"
    list_secrets: bool = False
    "Lists all secrets to eliminate non-existing keys upfront"


@configspec
class ConfigProvidersConfiguration(BaseConfiguration):
    # TODO: set to False in 2.0
    enable_airflow_secrets: bool = True
    enable_google_secrets: bool = False

    airflow_secrets: VaultProviderConfiguration = VaultProviderConfiguration(
        only_secrets=False, only_toml_fragments=False, list_secrets=False
    )
    google_secrets: VaultProviderConfiguration = VaultProviderConfiguration(
        only_secrets=True, only_toml_fragments=True, list_secrets=False
    )  # None  # dataclasses.field(default_factory=lambda: dict(only_secrets=True, only_toml_fragments=True, list_secrets=False))
    # VaultProviderConfiguration(only_secrets=True, only_toml_fragments=True, list_secrets=False)
    # always look in providers
    __section__: ClassVar[str] = known_sections.PROVIDERS


class ConfigProvidersContainer:
    """Injectable list of providers used by the configuration `resolve` module"""

    providers: List[ConfigProvider] = None
    context_provider: ConfigProvider = None

    def __init__(self, initial_providers: List[ConfigProvider]) -> None:
        super().__init__()
        # add default providers
        self.providers = initial_providers
        # ContextProvider will provide contexts when embedded in configurations
        self.context_provider = ContextProvider()

    def add_extras(self) -> None:
        """Adds extra providers. Extra providers may use initial providers when setting up"""
        for provider in _extra_providers():
            self[provider.name] = provider

    def __getitem__(self, name: str) -> ConfigProvider:
        try:
            return next(p for p in self.providers if p.name == name)
        except StopIteration:
            raise KeyError(name)

    def __setitem__(self, name: str, provider: ConfigProvider) -> None:
        idx = next((i for i, p in enumerate(self.providers) if p.name == name), -1)
        if idx == -1:
            self.providers.append(provider)
        else:
            self.providers[idx] = provider

    def __contains__(self, name: object) -> bool:
        try:
            self.__getitem__(name)  # type: ignore
            return True
        except KeyError:
            return False

    def add_provider(self, provider: ConfigProvider) -> None:
        if provider.name in self:
            raise DuplicateConfigProviderException(provider.name)
        self.providers.append(provider)


def _extra_providers() -> List[ConfigProvider]:
    """Providers that require initial providers to be instantiated as the are enabled via config"""
    from dlt.common.configuration.resolve import resolve_configuration

    providers_config = resolve_configuration(ConfigProvidersConfiguration())
    extra_providers = []
    if providers_config.enable_airflow_secrets:
        extra_providers.extend(_airflow_providers(providers_config.airflow_secrets))
    if providers_config.enable_google_secrets:
        extra_providers.append(_google_secrets_provider(providers_config.google_secrets))
    return extra_providers


def _google_secrets_provider(settings: VaultProviderConfiguration) -> ConfigProvider:
    from dlt.common.configuration.resolve import resolve_configuration
    from dlt.common.configuration.providers.google_secrets import GoogleSecretsProvider

    c = resolve_configuration(
        GcpServiceAccountCredentials(), sections=(known_sections.PROVIDERS, "google_secrets")
    )
    return GoogleSecretsProvider(c, **dict(settings))


def _airflow_providers(settings: VaultProviderConfiguration) -> List[ConfigProvider]:
    """Returns a list of configuration providers for an Airflow environment.

    This function attempts to import Airflow to determine whether it
    is running in an Airflow environment. If Airflow is not installed,
    an empty list is returned. If Airflow is installed, the function
    returns a list containing the Airflow providers.

    Depending on how DAG is defined this function may be called outside of task and
    task context will not be available. Still we want the provider to function so
    we just test if Airflow can be imported.
    """

    providers: List[ConfigProvider] = []

    try:
        # hide stdio. airflow typically dumps tons of warnings and deprecations to stdout and stderr
        with contextlib.redirect_stdout(io.StringIO()), contextlib.redirect_stderr(io.StringIO()):
            # try to get dlt secrets variable. many broken Airflow installations break here. in that case do not create
            from airflow.models import Variable, TaskInstance  # noqa
            from dlt.common.configuration.providers.airflow import AirflowSecretsTomlProvider

            # probe if Airflow variable containing all secrets is present
            from dlt.common.configuration.providers.vault import SECRETS_TOML_KEY

            secrets_toml_var = Variable.get(SECRETS_TOML_KEY, default_var=None)

            # providers can be returned - mind that AirflowSecretsTomlProvider() requests the variable above immediately
            providers = [AirflowSecretsTomlProvider(**dict(settings))]

            # check if we are in task context and provide more info
            from airflow.operators.python import get_current_context  # noqa

            ti: TaskInstance = get_current_context()["ti"]  # type: ignore[assignment,unused-ignore]

        # log outside of stderr/out redirect
        if secrets_toml_var is None:
            message = (
                f"Airflow variable '{SECRETS_TOML_KEY}' was not found. "
                + "This Airflow variable is a recommended place to hold the content of"
                " secrets.toml."
                + "If you do not use Airflow variables to hold dlt configuration or use variables"
                " with other names you can ignore this warning."
            )
            ti.log.warning(message)

    except Exception:
        # do not probe variables when not in task context
        pass

    # airflow not detected
    return providers



from typing import TYPE_CHECKING

# NOTE this warning is only trigger when type-checking.
# Move it outside the type checking block when closer to deprecation date.
if not TYPE_CHECKING:
    from dlt.common.warnings import DltDeprecationWarning

    DltDeprecationWarning(
        "Content from this module was moved to `dlt._dataset.utils`, which is internal.",
        since="1.16.0",
        expected_due="2.0.0",
    )

from dlt.dataset.utils import get_destination_client_initial_config, get_destination_clients



from typing import TYPE_CHECKING, Dict, Iterator, Optional, Sequence, List, Any

from dlt.common import logger
from dlt.common.destination.configuration import CsvFormatConfiguration
from dlt.common.destination import DestinationCapabilitiesContext
from dlt.common.destination.exceptions import (
    DestinationInvalidFileFormat,
)
from dlt.common.destination.client import (
    HasFollowupJobs,
    PreparedTableSchema,
    RunnableLoadJob,
    FollowupJobRequest,
    LoadJob,
)
from dlt.common.schema import TColumnSchema, TColumnHint, Schema
from dlt.common.schema.typing import TColumnType
from dlt.common.schema.utils import is_nullable_column
from dlt.common.storages.file_storage import FileStorage
from dlt.common.storages.load_storage import ParsedLoadJobFileName
from dlt.destinations._adbc_jobs import AdbcParquetCopyJob
from dlt.destinations.impl.postgres.configuration import PostgresClientConfiguration
from dlt.destinations.impl.postgres.factory import get_adbc_driver_location
from dlt.destinations.impl.postgres.sql_client import Psycopg2SqlClient
from dlt.destinations.insert_job_client import InsertValuesJobClient
from dlt.destinations.sql_client import SqlClientBase
from dlt.destinations.sql_jobs import SqlStagingReplaceFollowupJob

HINT_TO_POSTGRES_ATTR: Dict[TColumnHint, str] = {"unique": "UNIQUE"}


class PostgresStagingReplaceJob(SqlStagingReplaceFollowupJob):
    @classmethod
    def generate_sql(
        cls,
        table_chain: Sequence[PreparedTableSchema],
        sql_client: SqlClientBase[Any],
    ) -> List[str]:
        sql: List[str] = []
        for table in table_chain:
            with sql_client.with_staging_dataset():
                staging_table_name = sql_client.make_qualified_table_name(table["name"])
            table_name = sql_client.make_qualified_table_name(table["name"])
            sql.extend(
                (
                    f"DROP TABLE IF EXISTS {table_name}",
                    (
                        f"ALTER TABLE {staging_table_name} SET SCHEMA"
                        f" {sql_client.fully_qualified_dataset_name()}"
                    ),
                    f"CREATE TABLE {staging_table_name} (like {table_name} including all)",
                )
            )
        return sql


class PostgresParquetCopyJob(AdbcParquetCopyJob):
    if TYPE_CHECKING:
        from adbc_driver_manager.dbapi import Connection

    def _connect(self) -> "Connection":
        from adbc_driver_manager import dbapi

        self._config = self._job_client.config
        return dbapi.connect(
            driver=get_adbc_driver_location(),
            db_kwargs={"uri": self._config.credentials.to_native_representation()},
        )


class PostgresCsvCopyJob(RunnableLoadJob, HasFollowupJobs):
    def __init__(self, file_path: str) -> None:
        super().__init__(file_path)
        self._job_client: PostgresClient = None

    def run(self) -> None:
        self._config = self._job_client.config
        sql_client = self._job_client.sql_client
        csv_format = self._config.csv_format or CsvFormatConfiguration()
        table_name = self.load_table_name
        sep = csv_format.delimiter
        if csv_format.on_error_continue:
            logger.warning(
                f"When processing {self._file_path} on table {table_name} Postgres csv reader does"
                " not support on_error_continue"
            )

        with FileStorage.open_zipsafe_ro(self._file_path, "rb") as f:
            if csv_format.include_header:
                # all headers in first line
                headers_row: str = f.readline().decode(csv_format.encoding).strip()
                split_headers = headers_row.split(sep)
            else:
                # read first row to figure out the headers
                split_first_row: str = f.readline().decode(csv_format.encoding).strip().split(sep)
                split_headers = list(self._job_client.schema.get_table_columns(table_name).keys())
                if len(split_first_row) > len(split_headers):
                    raise DestinationInvalidFileFormat(
                        "postgres",
                        "csv",
                        self._file_path,
                        f"First row `{split_first_row}` has more columns than header"
                        f" `{split_headers}` in table `{table_name}`",
                    )
                if len(split_first_row) < len(split_headers):
                    logger.warning(
                        f"First row {split_first_row} has less columns than header"
                        f" `{split_headers}` in table `{table_name}`. We will not load data to"
                        " superfluous columns."
                    )
                    split_headers = split_headers[: len(split_first_row)]
                # stream the first row again
                f.seek(0)

            # normalized and quoted headers
            split_headers = [
                sql_client.escape_column_name(h.strip('"'), quote=True) for h in split_headers
            ]
            split_null_headers = []
            split_columns = []
            # detect columns with NULL to use in FORCE NULL
            # detect headers that are not in columns
            for col in self._job_client.schema.get_table_columns(table_name).values():
                norm_col = sql_client.escape_column_name(col["name"], quote=True)
                split_columns.append(norm_col)
                if norm_col in split_headers and is_nullable_column(col):
                    split_null_headers.append(norm_col)
            if split_unknown_headers := set(split_headers).difference(split_columns):
                raise DestinationInvalidFileFormat(
                    "postgres",
                    "csv",
                    self._file_path,
                    f"Following headers {split_unknown_headers} cannot be matched to columns"
                    f" {split_columns} of table {table_name}.",
                )

            # use comma to join
            headers = ",".join(split_headers)
            if split_null_headers:
                null_headers = f"FORCE_NULL({','.join(split_null_headers)}),"
            else:
                null_headers = ""

            qualified_table_name = sql_client.make_qualified_table_name(table_name)
            copy_sql = (
                f"COPY {qualified_table_name} ({headers}) FROM STDIN WITH (FORMAT CSV, DELIMITER"
                f" '{sep}', NULL '', {null_headers} ENCODING '{csv_format.encoding}')"
            )
            with sql_client.begin_transaction():
                with sql_client.native_connection.cursor() as cursor:
                    cursor.copy_expert(copy_sql, f, size=8192)


class PostgresClient(InsertValuesJobClient):
    def __init__(
        self,
        schema: Schema,
        config: PostgresClientConfiguration,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        dataset_name, staging_dataset_name = InsertValuesJobClient.create_dataset_names(
            schema, config
        )
        sql_client = Psycopg2SqlClient(
            dataset_name,
            staging_dataset_name,
            config.credentials,
            capabilities,
        )
        super().__init__(schema, config, sql_client)
        self.config: PostgresClientConfiguration = config
        self.sql_client: Psycopg2SqlClient = sql_client
        self.active_hints = HINT_TO_POSTGRES_ATTR if self.config.create_indexes else {}
        self.type_mapper = self.capabilities.get_type_mapper()

    def create_load_job(
        self, table: PreparedTableSchema, file_path: str, load_id: str, restore: bool = False
    ) -> LoadJob:
        job = super().create_load_job(table, file_path, load_id, restore)
        if not job:
            parsed_file = ParsedLoadJobFileName.parse(file_path)
            if parsed_file.file_format == "csv":
                job = PostgresCsvCopyJob(file_path)
            elif parsed_file.file_format == "parquet":
                job = PostgresParquetCopyJob(file_path)
        return job

    def _create_replace_followup_jobs(
        self, table_chain: Sequence[PreparedTableSchema]
    ) -> List[FollowupJobRequest]:
        root_table = table_chain[0]
        if root_table["x-replace-strategy"] == "staging-optimized":  # type: ignore[typeddict-item]
            return [PostgresStagingReplaceJob.from_table_chain(table_chain, self.sql_client)]
        return super()._create_replace_followup_jobs(table_chain)

    def _from_db_type(
        self, pq_t: str, precision: Optional[int], scale: Optional[int]
    ) -> TColumnType:
        return self.type_mapper.from_destination_type(pq_t, precision, scale)



import functools
import os
from tempfile import gettempdir
from typing import Any, Callable, Dict, List, Literal, Optional, Sequence, Tuple

from tenacity import (
    retry_if_exception,
    wait_exponential,
    stop_after_attempt,
    Retrying,
    RetryCallState,
)

from dlt.common.known_env import DLT_DATA_DIR, DLT_PROJECT_DIR, DLT_LOCAL_DIR
from dlt.common.exceptions import MissingDependencyException, ValueErrorWithKnownValues

try:
    from airflow.configuration import conf
    from airflow.models import TaskInstance, BaseOperator
    from airflow.utils.task_group import TaskGroup
    from airflow.operators.empty import EmptyOperator
    from airflow.operators.python import PythonOperator, get_current_context
except ModuleNotFoundError:
    raise MissingDependencyException("Airflow", ["apache-airflow>=2.5"])


import dlt
from dlt.common import logger
from dlt.common.pendulum import pendulum
from dlt.common.runtime.telemetry import with_telemetry

from dlt.common.destination import TLoaderFileFormat
from dlt.common.schema.typing import TWriteDispositionConfig, TSchemaContract

from dlt.common.utils import uniq_id
from dlt.common.normalizers.naming.snake_case import NamingConvention as SnakeCaseNamingConvention
from dlt.common.configuration.container import Container
from dlt.common.configuration.specs.pluggable_run_context import PluggableRunContext
from dlt.common.runtime.collector import NULL_COLLECTOR

from dlt.extract import DltSource
from dlt.pipeline.helpers import retry_load
from dlt.pipeline.pipeline import Pipeline
from dlt.pipeline.progress import log
from dlt.pipeline.typing import TPipelineStep


DEFAULT_RETRY_NO_RETRY = Retrying(stop=stop_after_attempt(1), reraise=True)
DEFAULT_RETRY_BACKOFF = Retrying(
    stop=stop_after_attempt(5), wait=wait_exponential(multiplier=1.5, min=4, max=10), reraise=True
)


class PipelineTasksGroup(TaskGroup):
    """
    Represents a dlt Airflow pipeline task group.
    """

    def __init__(
        self,
        pipeline_name: str,
        use_data_folder: bool = False,
        local_data_folder: str = None,
        use_task_logger: bool = True,
        log_progress_period: float = 30.0,
        buffer_max_items: int = 1000,
        retry_policy: Retrying = DEFAULT_RETRY_NO_RETRY,
        retry_pipeline_steps: Sequence[TPipelineStep] = ("load",),
        abort_task_if_any_job_failed: bool = True,
        wipe_local_data: bool = True,
        save_load_info: bool = False,
        save_trace_info: bool = False,
        **kwargs: Any,
    ) -> None:
        """Creates a task group to which you can add pipeline runs

        The run environment is prepared as follows
        - the .dlt folder (the project folder) is searched under `dags` as configured by Airflow
        - the data folder where pipelines are stored is always unique

        The `data_folder` is available in certain Airflow deployments. In case of Composer, it is a location on the gcs bucket. `use_data_folder` is disabled and should be
        enabled only when needed. The operations on bucket are non-atomic and way slower than on local storage and should be avoided.

        `abort_task_if_any_job_failed` will abort the other dlt loading jobs and fail the Airflow task in any of the jobs failed. See https://dlthub.com/docs/running-in-production/running#handle-exceptions-failed-jobs-and-retry-the-pipeline.

        The load info and trace info can be optionally saved to the destination. See https://dlthub.com/docs/running-in-production/running#inspect-and-save-the-load-info-and-trace

        Args:
            pipeline_name (str): Name of the task group
            use_data_folder (bool, optional): If well defined 'data' folder is present it will be used. Currently only data folder on Composer is supported. Defaults to False.
            local_data_folder (str, optional): Path to a local folder on worker machine to where to store data. Used if local_data_folder is False or there's not well defined data folder. Defaults to gettempdir.
            use_task_logger (bool, optional): Will redirect dlt logger into task logger. Defaults to True.
            log_progress_period (float, optional): If progress is not configured for a pipeline, the `log` progress is used with a given period. Set 0 to disable. Defaults to 30.0.
            buffer_max_items (int, optional): Maximum number of buffered items. Use 0 to keep dlt built-in limit. Defaults to 1000.
            retry_policy (_type_, optional): Tenacity retry policy. Defaults to no retry.
            retry_pipeline_steps (Sequence[TPipelineStep], optional): Which pipeline steps are eligible for retry. Defaults to ("load", ).
            wipe_local_data (bool, optional): Will wipe all the data created by pipeline, also in case of exception. Defaults to True.
            save_load_info (bool, optional): Will save extensive load info to the destination. Defaults to False.
            save_trace_info (bool, optional): Will save trace info to the destination. Defaults to False.
        """

        super().__init__(group_id=pipeline_name, **kwargs)
        self._used_names: Dict[str, Any] = {}
        self.use_task_logger = use_task_logger
        self.log_progress_period = log_progress_period
        self.buffer_max_items = buffer_max_items
        self.retry_policy = retry_policy
        self.retry_pipeline_steps = retry_pipeline_steps
        self.abort_task_if_any_job_failed = abort_task_if_any_job_failed
        self.wipe_local_data = wipe_local_data
        self.save_load_info = save_load_info
        self.save_trace_info = save_trace_info

        # reload providers so config.toml in dags folder is included
        dags_folder = conf.get("core", "dags_folder")

        # TODO: use injectable Airflow run context
        # set the dlt project folder to dags
        os.environ[DLT_PROJECT_DIR] = dags_folder

        # check if /data mount is available
        if use_data_folder and os.path.exists("/home/airflow/gcs/data"):
            data_dir = os.path.join("/home/airflow/gcs/data", f"dlt_{uniq_id(8)}")
        else:
            # create random path
            data_dir = os.path.join(local_data_folder or gettempdir(), f"dlt_{uniq_id(8)}")
        os.environ[DLT_DATA_DIR] = data_dir
        # also keep all local files created by destinations in data_dir
        os.environ[DLT_LOCAL_DIR] = data_dir

        # reload config providers (TODO: inject Airflow run context)
        if PluggableRunContext in Container():
            Container()[PluggableRunContext].reload_providers()

    def _task_name(self, pipeline: Pipeline, data: Any) -> str:
        """Generate a task name.

        Args:
            pipeline (Pipeline): The pipeline to run.
            data (Any): The data to run the pipeline with.

        Returns:
            str: The name of the task.
        """
        task_name = pipeline.pipeline_name

        if isinstance(data, DltSource):
            resource_names = list(data.selected_resources.keys())
            task_name = data.name + "_" + "-".join(resource_names[:4])

            if len(resource_names) > 4:
                task_name += f"-{len(resource_names)-4}-more"

            num = self._used_names.setdefault(task_name, 0)
            self._used_names[task_name] = num + 1

            if num:
                task_name += f"-{num + 1}"

        return task_name

    def run(
        self,
        pipeline: Pipeline,
        data: Any,
        table_name: str = None,
        write_disposition: TWriteDispositionConfig = None,
        loader_file_format: TLoaderFileFormat = None,
        schema_contract: TSchemaContract = None,
        pipeline_name: str = None,
        on_before_run: Callable[[], None] = None,
        **kwargs: Any,
    ) -> PythonOperator:
        """
        Create a task to run the given pipeline with the
        given data in Airflow.

        Args:
            pipeline (Pipeline): The pipeline to run
            data (Any):
                The data to run the pipeline with. If a non-resource
                callable given, it's evaluated during the DAG execution,
                right before the actual pipeline run.
                NOTE: If `on_before_run` is provided, first `on_before_run`
                      is evaluated, and then callable `data`.
            table_name (str, optional): The name of the table to
                which the data should be loaded within the `dataset`.
            write_disposition (TWriteDispositionConfig, optional): Same as
                in `run` command.
            loader_file_format (TLoaderFileFormat, optional):
                The file format the loader will use to create the
                load package.
            schema_contract (TSchemaContract, optional): On override
                for the schema contract settings, this will replace
                the schema contract settings for all tables in the schema.
            pipeline_name (str, optional): The name of the derived pipeline.
            on_before_run (Callable, optional): A callable to be
                executed right before the actual pipeline run.

        Returns:
            PythonOperator: Airflow task instance.
        """
        f = functools.partial(
            self._run,
            pipeline,
            data,
            table_name=table_name,
            write_disposition=write_disposition,
            loader_file_format=loader_file_format,
            schema_contract=schema_contract,
            pipeline_name=pipeline_name,
            on_before_run=on_before_run,
        )
        return PythonOperator(task_id=self._task_name(pipeline, data), python_callable=f, **kwargs)

    def _run(
        self,
        pipeline: Pipeline,
        data: Any,
        table_name: str = None,
        write_disposition: TWriteDispositionConfig = None,
        loader_file_format: TLoaderFileFormat = None,
        schema_contract: TSchemaContract = None,
        pipeline_name: str = None,
        on_before_run: Callable[[], None] = None,
    ) -> None:
        """Run the given pipeline with the given data.

        Args:
            pipeline (Pipeline): The pipeline to run
            data (Any):
                The data to run the pipeline with. If a non-resource
                callable given, it's evaluated during the DAG execution,
                right before the actual pipeline run.
                NOTE: If `on_before_run` is provided, first `on_before_run`
                      is evaluated, and then callable `data`.
            table_name (str, optional): The name of the
                table to which the data should be loaded
                within the `dataset`.
            write_disposition (TWriteDispositionConfig, optional):
                Same as in `run` command.
            loader_file_format (TLoaderFileFormat, optional):
                The file format the loader will use to create
                the load package.
            schema_contract (TSchemaContract, optional): On
                override for the schema contract settings,
                this will replace the schema contract settings
                for all tables in the schema.
            pipeline_name (str, optional): The name of the
                derived pipeline.
            on_before_run (Callable, optional): A callable
                to be executed right before the actual pipeline run.
        """
        # activate pipeline
        pipeline.activate()
        # drop local data
        task_pipeline = pipeline.drop(pipeline_name=pipeline_name)

        # use task logger
        if self.use_task_logger:
            ti: TaskInstance = get_current_context()["ti"]  # type: ignore[assignment,unused-ignore]
            logger.LOGGER = ti.log

        # set global number of buffered items
        if dlt.config.get("data_writer.buffer_max_items") is None and self.buffer_max_items > 0:
            dlt.config["data_writer.buffer_max_items"] = self.buffer_max_items
            logger.info(f"Set data_writer.buffer_max_items to {self.buffer_max_items}")

        if self.abort_task_if_any_job_failed is not None:
            dlt.config["load.raise_on_failed_jobs"] = self.abort_task_if_any_job_failed
            logger.info(
                f"Set load.abort_task_if_any_job_failed to {self.abort_task_if_any_job_failed}"
            )

        if self.log_progress_period > 0 and task_pipeline.collector == NULL_COLLECTOR:
            task_pipeline.collector = log(log_period=self.log_progress_period, logger=logger.LOGGER)
            logger.info(f"Enabled log progress with period {self.log_progress_period}")

        logger.info(f"Pipeline data in {task_pipeline.working_dir}")

        def log_after_attempt(retry_state: RetryCallState) -> None:
            if not retry_state.retry_object.stop(retry_state):
                logger.error(
                    "Retrying pipeline run due to exception: %s",
                    retry_state.outcome.exception(),
                )

        try:
            if on_before_run is not None:
                on_before_run()

            if callable(data):
                data = data()

            # retry with given policy on selected pipeline steps
            for attempt in self.retry_policy.copy(
                retry=retry_if_exception(
                    retry_load(retry_on_pipeline_steps=self.retry_pipeline_steps)
                ),
                after=log_after_attempt,
            ):
                with attempt:
                    logger.info(
                        "Running the pipeline, attempt=%s" % attempt.retry_state.attempt_number
                    )
                    load_info = task_pipeline.run(
                        data,
                        table_name=table_name,
                        write_disposition=write_disposition,
                        loader_file_format=loader_file_format,
                        schema_contract=schema_contract,
                    )
                    logger.info(str(load_info))
                    # save load and trace
                    if self.save_load_info:
                        logger.info("Saving the load info in the destination")
                        task_pipeline.run(
                            [load_info],
                            table_name="_load_info",
                            loader_file_format=loader_file_format,
                        )
                    if self.save_trace_info:
                        logger.info("Saving the trace in the destination")
                        task_pipeline.run(
                            [task_pipeline.last_trace],
                            table_name="_trace",
                            loader_file_format=loader_file_format,
                        )

        finally:
            # always completely wipe out pipeline folder, in case of success and failure
            if self.wipe_local_data:
                logger.info(f"Removing folder {pipeline.working_dir}")
                task_pipeline._wipe_working_folder()

    @with_telemetry("helper", "airflow_add_run", False, "decompose")
    def add_run(
        self,
        pipeline: Pipeline,
        data: Any,
        *,
        decompose: Literal["none", "serialize", "parallel", "parallel-isolated"] = "none",
        table_name: str = None,
        write_disposition: TWriteDispositionConfig = None,
        loader_file_format: TLoaderFileFormat = None,
        schema_contract: TSchemaContract = None,
        on_before_run: Callable[[], None] = None,
        **kwargs: Any,
    ) -> List[BaseOperator]:
        """Creates a task or a group of tasks to run `data` with `pipeline`

        Creates an Airflow task that extracts, normalizes and loads `data` with the passed pipeline instance `pipeline`. If `data` is a source
        and `decompose` is `serialize` it will decompose the source into disjoint connected components (isolated group of resources) and execute them
        one after another as separate Airflow tasks. The decomposition makes sure that each resource or transformer is extracted only once. It preserves
        the order of resources declared in the source when creating graph of tasks.

        The `kwargs` are passed as arguments to all Airflow task instances created.

        Args:
            pipeline (Pipeline): An instance of pipeline used to run the source
            data (Any):
                Any data supported by `run` method of the pipeline.
                If a non-resource callable given, it's called before
                the load to get the data.
            decompose (Literal["none", "serialize", "parallel"], optional):
                A source decomposition strategy into Airflow tasks:
                    none - no decomposition, default value.
                    serialize - decompose the source into a sequence of Airflow tasks.
                    parallel - decompose the source into a parallel Airflow task group,
                               except the first resource must be completed first.
                               All tasks that are run in parallel share the same pipeline state.
                               If two of them modify the state, part of state may be lost
                    parallel-isolated - decompose the source into a parallel Airflow task group.
                               with the same exception as above. All task have separate pipeline
                               state (via separate pipeline name) but share the same dataset,
                               schemas and tables.
                NOTE: The first component of the source in both parallel models is done first,
                      after that the rest are executed in parallel to each other.
                NOTE: In case the SequentialExecutor is used by Airflow, the tasks
                      will remain sequential despite 'parallel' or 'parallel-isolated' mode.
                      Use another executor (e.g. CeleryExecutor) to make tasks parallel!

                Parallel tasks are executed in different pipelines, all derived from the original
                one, but with the state isolated from each other.
            table_name: (str): The name of the table to which the data should be loaded within the `dataset`
            write_disposition (TWriteDispositionConfig, optional): Same as in `run` command. Defaults to None.
            loader_file_format (Literal["jsonl", "insert_values", "parquet"], optional): The file format the loader will use to create the load package.
                Not all file_formats are compatible with all destinations. Defaults to the preferred file format of the selected destination.
            schema_contract (TSchemaContract, optional): On override for the schema contract settings,
                this will replace the schema contract settings for all tables in the schema. Defaults to None.
            on_before_run (Callable, optional):
                A callable to be executed right before the actual pipeline run.

        Returns:
            Any: Airflow tasks created in order of creation.
        """

        # make sure that pipeline was created after dag was initialized
        if not pipeline.pipelines_dir.startswith(os.environ[DLT_DATA_DIR]):
            raise ValueError(
                "Please create your Pipeline instance after AirflowTasks are created. The dlt"
                f" pipelines directory {pipeline.pipelines_dir} is not set correctly"
                f" ({os.environ[DLT_DATA_DIR]} expected)."
            )

        with self:
            # use factory function to make a task, in order to parametrize it
            # passing arguments to task function (_run) is serializing
            # them and running template engine on them
            def make_task(pipeline: Pipeline, data: Any, name: str = None) -> BaseOperator:
                f = functools.partial(
                    self._run,
                    pipeline,
                    data,
                    table_name=table_name,
                    write_disposition=write_disposition,
                    loader_file_format=loader_file_format,
                    schema_contract=schema_contract,
                    pipeline_name=name,
                    on_before_run=on_before_run,
                )
                return PythonOperator(
                    task_id=self._task_name(pipeline, data), python_callable=f, **kwargs
                )

            if decompose == "none":
                # run pipeline as single task
                return [make_task(pipeline, data)]
            elif decompose == "serialize":
                if not isinstance(data, DltSource):
                    raise ValueError("Can only decompose dlt sources")
                if pipeline.dev_mode:
                    raise ValueError("Cannot decompose pipelines with `dev_mode=True`")
                # serialize tasks
                tasks = []
                pt = None
                for source in data.decompose("scc"):
                    nt = make_task(pipeline, source)
                    tasks.append(nt)
                    if pt is not None:
                        pt >> nt
                    pt = nt
                return tasks
            elif decompose == "parallel":
                if not isinstance(data, DltSource):
                    raise ValueError("Can only decompose dlt sources")

                if pipeline.dev_mode:
                    raise ValueError("Cannot decompose pipelines with `dev_mode=True`")

                tasks = []
                sources = data.decompose("scc")
                t_name = self._task_name(pipeline, data)
                start = make_task(pipeline, sources[0])

                # parallel tasks
                for source in sources[1:]:
                    for resource in source.resources.values():
                        if resource.incremental:
                            logger.warn(
                                f"The resource {resource.name} in task {t_name} "
                                "is using incremental loading and may modify the "
                                "state. Resources that modify the state should not "
                                "run in parallel within the single pipeline as the "
                                "state will not be correctly merged. Please use "
                                "'serialize' or 'parallel-isolated' modes instead."
                            )
                            break

                    tasks.append(make_task(pipeline, source))

                end = EmptyOperator(task_id=f"{t_name}_end")

                if tasks:
                    start >> tasks >> end
                    return [start] + tasks + [end]

                start >> end
                return [start, end]
            elif decompose == "parallel-isolated":
                if not isinstance(data, DltSource):
                    raise ValueError("Can only decompose dlt sources")

                if pipeline.dev_mode:
                    raise ValueError("Cannot decompose pipelines with `dev_mode=True`")

                # parallel tasks
                tasks = []
                naming = SnakeCaseNamingConvention()
                sources = data.decompose("scc")
                start = make_task(
                    pipeline,
                    sources[0],
                    naming.normalize_identifier(self._task_name(pipeline, sources[0])),
                )

                # parallel tasks
                for source in sources[1:]:
                    # name pipeline the same as task
                    new_pipeline_name = naming.normalize_identifier(
                        self._task_name(pipeline, source)
                    )
                    tasks.append(make_task(pipeline, source, new_pipeline_name))

                t_name = self._task_name(pipeline, data)
                end = EmptyOperator(task_id=f"{t_name}_end")

                if tasks:
                    start >> tasks >> end
                    return [start] + tasks + [end]

                start >> end
                return [start, end]
            else:
                raise ValueErrorWithKnownValues(
                    "decompose", decompose, ["none", "serialize", "parallel", "parallel-isolated"]
                )


def airflow_get_execution_dates() -> Tuple[pendulum.DateTime, Optional[pendulum.DateTime]]:
    # prefer logging to task logger
    try:
        from airflow.operators.python import get_current_context  # noqa

        context = get_current_context()
        return context["data_interval_start"], context["data_interval_end"]
    except Exception:
        return None, None



from typing import NamedTuple, Sequence

from dlt.common.destination import TLoaderFileFormat
from dlt.common.exceptions import DltException


class DataWriterException(DltException):
    pass


class InvalidFileNameTemplateException(DataWriterException, ValueError):
    def __init__(self, file_name_template: str):
        self.file_name_template = file_name_template
        super().__init__(
            f"Wrong file name template `{file_name_template}`. File name template must contain"
            " exactly one %s formatter"
        )


class BufferedDataWriterClosed(DataWriterException):
    def __init__(self, file_name: str):
        self.file_name = file_name
        super().__init__(f"Writer with recent file name `{file_name}` is already closed")


class FileImportNotFound(DataWriterException, FileNotFoundError):
    def __init__(self, import_file_path: str, local_file_path: str) -> None:
        self.import_file_path = import_file_path
        self.local_file_path = local_file_path
        super().__init__(
            f"Attempt to import non existing file `{import_file_path}` into extract storage file"
            f" `{local_file_path}`"
        )


class DestinationCapabilitiesRequired(DataWriterException, ValueError):
    def __init__(self, file_format: TLoaderFileFormat):
        self.file_format = file_format
        super().__init__(
            f"Writer for `{file_format=:}` requires destination capabilities which were not"
            " provided."
        )


class DataWriterNotFound(DataWriterException):
    pass


class FileFormatForItemFormatNotFound(DataWriterNotFound):
    def __init__(self, file_format: TLoaderFileFormat, data_item_format: str):
        self.file_format = file_format
        self.data_item_format = data_item_format
        super().__init__(
            f"Can't find a file writer for `{file_format=:}` and item format `{data_item_format=:}`"
        )


class FileSpecNotFound(KeyError, DataWriterNotFound):
    def __init__(self, file_format: TLoaderFileFormat, data_item_format: str, spec: NamedTuple):
        self.file_format = file_format
        self.data_item_format = data_item_format
        super().__init__(
            f"Can't find a file writer for spec with `{file_format=:}` and `{data_item_format=:}`"
            f" where the full spec is `{spec}`"
        )


class SpecLookupFailed(DataWriterNotFound):
    def __init__(
        self,
        data_item_format: str,
        possible_file_formats: Sequence[TLoaderFileFormat],
        file_format: TLoaderFileFormat,
    ):
        self.file_format = file_format
        self.possible_file_formats = possible_file_formats
        self.data_item_format = data_item_format
        super().__init__(
            f"Failed to find file writer for {data_item_format=:} among file formats"
            f" {possible_file_formats=:}. The preferred file format was `{file_format=:}`."
        )


class InvalidDataItem(DataWriterException):
    def __init__(self, file_format: TLoaderFileFormat, data_item_format: str, details: str):
        self.file_format = file_format
        self.data_item_format = data_item_format
        super().__init__(
            f"A data item of type {data_item_format=:} cannot be written as `{file_format}:"
            f" {details}`"
        )



from dlt.common import logger
from dlt.common.arithmetics import Decimal
from dlt.common.wei import Wei
from dlt.common.pendulum import pendulum
from dlt.common.json import json
from dlt.common.runtime.signals import sleep

__all__ = ["Decimal", "Wei", "pendulum", "json", "sleep", "logger"]
